(*
 * Calls and encodings of run-time system functions.
 *)

open Map

type tag  = 
    TAG_TEST_EQUALITY 
  | TAG_DUMMY
  | TAG_GETSITE

module CallsMap =
  Map.Make(struct
    type t = string
    let compare = compare
  end)

let c = CallsMap.empty
let c = CallsMap.add "test_equality " TAG_TEST_EQUALITY c
let c = CallsMap.add "dummy "         TAG_DUMMY c
let c = CallsMap.add "getSite "       TAG_GETSITE c 
    
let yieldCall i l =
  let key i = match i with 
    (i::_) -> i 
  |  _  ->  Error.bug "Calls.yieldCall: Expected (i::_)"
  in
  let i = key i in
  try match CallsMap.find i c with
    TAG_TEST_EQUALITY ->
      (match l with  (* warning: some parameters which might receive 
			correct arguments (i.e. comparable names) are not 
			recognised below, e.g. projections of records *)
	
	Inter.DYNAMIC(x)::Inter.DYNAMIC(y)::[] -> Inter.TEST_EQUALITY(x,y)
      | Inter.STATIC(x)::Inter.STATIC(y)::[] -> Inter.TEST_EQUALITY(x,y)
      | _ -> raise (Invalid_argument i))
  | TAG_DUMMY -> Inter.DUMMY
  | TAG_GETSITE -> Inter.GETSITE
  with Not_found -> Error.errAt Error.progName 
      ("Unknown internal function: \""^i^"\"")
      
let caml_printTag = function
    TAG_TEST_EQUALITY -> Format.print_string "TAG_TEST_EQUALITY"
  | TAG_DUMMY -> Format.print_string "TAG_DUMMY"
  | TAG_GETSITE -> Format.print_string "TAG_GETSITE"

