
let iterateInt start finish step f res =
  if step = 0 then
    Error.bug "Misc.iterateInt"
  else if step > 0 then 
    let rec iterateUp index res =
      if index > finish then res
      else iterateUp (index + step) (f index res)
    in
      iterateUp start res
  else
    let rec iterateDown index res =
      if index < finish then res
      else iterateDown (index + step) (f index res)
    in
      iterateDown start res

module StringSet =
  Set.Make(struct
    type t = string
    let compare = compare
  end)

(* old, working with Ocaml-1.05 
module type S =
  sig
    type key
    type 'a t
    val empty: 'a t
    val add: key -> 'a -> 'a t -> 'a t
    val find: key -> 'a t -> 'a
    val remove: key -> 'a t -> 'a t
    val iter: (key -> 'a -> 'b) -> 'a t -> unit
    val fold: (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
    val union : 'a t -> 'a t -> 'a t
    val map : (key -> 'a -> 'b) -> 'a t -> 'b t
    val format :
      string -> string -> string -> (key -> 'a -> unit) -> 'a t -> unit
  end

module Make (Ord:Map.OrderedType) : (S with type key = Ord.t) =
  struct
    module M = Map.Make(Ord)
    type key = Ord.t
    type 'a t = 'a M.t
    let empty = M.empty
    let add = M.add
    let find = M.find
    let remove = M.remove
    let iter = M.iter
    let fold = M.fold
    let union m1 m2 = M.fold M.add m2 m1
    let map f m = M.fold (fun k e m -> M.add k (f k e) m) m M.empty

    let format start finish sep p m =
      let rec pr k e needSep =
	if needSep then
	  (Format.print_string sep; Format.print_cut(); p k e; true)
	else
	  (p k e; true)
    in
      Format.print_string start;
      Format.open_hovbox 0; M.fold pr m false;
      Format.print_string finish;
      Format.close_box()
  end
*)

(* old, working with Ocaml-3.0 ? 

module type S =
  sig
    type key
    type +'a t
    val empty: 'a t
    val add: key -> 'a -> 'a t -> 'a t
    val find: key -> 'a t -> 'a
    val remove: key -> 'a t -> 'a t
    val mem : key -> 'a t -> bool
    val iter : (key -> 'a -> unit) -> 'a t -> unit
    val fold: (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
    val union : 'a t -> 'a t -> 'a t
    val map : ('a -> 'b) -> 'a t -> 'b t
(*    val mapi : f:(key -> 'a -> 'b) -> 'a t -> 'b t *)
    val mapi : (key -> 'a -> 'b) -> 'a t -> 'b t
    val mymap : (key -> 'a -> 'b) -> 'a t -> 'b t
    val format :
      string -> string -> string -> (key -> 'a -> unit) -> 'a t -> unit
  end

module Make (Ord:Map.OrderedType) : (S with type key = Ord.t) =
  struct
    module M = Map.Make(Ord)
    type key = Ord.t
    type 'a t = 'a M.t
    let empty = M.empty
    let add = M.add
    let find = M.find
    let remove = M.remove
    let mem = M.mem
    let iter = M.iter
    let fold = M.fold
    let union m1 m2 = M.fold M.add m2 m1
    let map = M.map
    let mapi = M.mapi
    let mymap f m = M.fold (fun k e m -> M.add k (f k e) m) m M.empty

    let format start finish sep p m =
      let rec pr k e needSep =
	if needSep then
	  (Format.print_string sep; Format.print_cut(); p k e; true)
	else
	  (p k e; true)
    in
      Format.print_string start;
      Format.open_hovbox 0; let _ = M.fold pr m false in ();
      Format.print_string finish;
      Format.close_box()
  end

*)

module type S =
  sig
    type key
    type (+'a) t
    val empty : 'a t
    val is_empty : 'a t -> bool
    val mem : key -> 'a t -> bool
    val add : key -> 'a -> 'a t -> 'a t
    val update : key -> ('a option -> 'a option) -> 'a t -> 'a t
    val singleton : key -> 'a -> 'a t
    val remove : key -> 'a t -> 'a t
    val merge : (key -> 'a option -> 'b option -> 'c option) -> 'a t -> 'b t -> 'c t
	(*
	  val union : 'a t -> 'a t -> 'a t  (* reimplemented *)
	 *)
    val union : (key -> 'a -> 'a -> 'a option) -> 'a t -> 'a t -> 'a t
    val compare : ('a -> 'a -> int) -> 'a t -> 'a t -> int
    val equal : ('a -> 'a -> bool) -> 'a t -> 'a t -> bool
    val iter : (key -> 'a -> unit) -> 'a t -> unit
    val fold : (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
    val for_all : (key -> 'a -> bool) -> 'a t -> bool
    val exists : (key -> 'a -> bool) -> 'a t -> bool
    val filter : (key -> 'a -> bool) -> 'a t -> 'a t
    val partition : (key -> 'a -> bool) -> 'a t -> 'a t * 'a t
    val cardinal : 'a t -> int
    val bindings : 'a t -> (key * 'a) list
    val min_binding : 'a t -> key * 'a
    val min_binding_opt : 'a t -> (key * 'a) option
    val max_binding : 'a t -> key * 'a
    val max_binding_opt : 'a t -> (key * 'a) option
    val choose : 'a t -> key * 'a
    val choose_opt : 'a t -> (key * 'a) option
    val split : key -> 'a t -> 'a t * 'a option * 'a t	      
    val find : key -> 'a t -> 'a
    val find_opt : key -> 'a t -> 'a option
    val find_first : (key -> bool) -> 'a t -> key * 'a
    val find_first_opt : (key -> bool) -> 'a t -> (key * 'a) option
    val find_last : (key -> bool) -> 'a t -> key * 'a
    val find_last_opt : (key -> bool) -> 'a t -> (key * 'a) option
    val map : ('a -> 'b) -> 'a t -> 'b t
    val mapi : (key -> 'a -> 'b) -> 'a t -> 'b t
    val to_seq : 'a t -> (key * 'a) Stdlib.Seq.t
    val to_seq_from : key -> 'a t -> (key * 'a) Stdlib.Seq.t
    val add_seq : (key * 'a) Stdlib.Seq.t -> 'a t -> 'a t
    val of_seq : (key * 'a) Stdlib.Seq.t -> 'a t
	(* added new methods *)
    val myunion : 'a t -> 'a t -> 'a t
    val mymap : (key -> 'a -> 'b) -> 'a t -> 'b t
    val format :
      string -> string -> string -> (key -> 'a -> unit) -> 'a t -> unit 
  end

module Make (Ord:Map.OrderedType) : (S with type key = Ord.t) =
  struct
    module M = Map.Make(Ord)
    type key = Ord.t
    type 'a t = 'a M.t
    let empty = M.empty
    let is_empty = M.is_empty
    let mem = M.mem
    let add = M.add
    let update = M.update
    let singleton = M.singleton
    let remove = M.remove
    let merge = M.merge
(*
    let union m1 m2 = M.fold M.add m2 m1 (* reimplemented! *)
 *)
    let union = M.union
    let compare = M.compare
    let equal = M.equal
    let iter = M.iter
    let fold = M.fold
    let for_all = M.for_all
    let exists = M.exists
    let filter = M.filter
    let partition = M.partition
    let cardinal = M.cardinal
    let bindings = M.bindings
    let min_binding = M.min_binding
    let min_binding_opt = M.min_binding_opt
    let max_binding = M.max_binding
    let max_binding_opt = M.max_binding_opt
    let choose = M.choose
    let choose_opt = M.choose_opt
    let split = M.split
    let find = M.find
    let find_opt = M.find_opt
    let find_first = M.find_first
    let find_first_opt = M.find_first_opt
    let find_last = M.find_last
    let find_last_opt = M.find_last_opt
    let map = M.map
    let mapi = M.mapi
    let to_seq = M.to_seq
    let to_seq_from = M.to_seq_from
    let add_seq = M.add_seq 
    let of_seq = M.of_seq
	(* added new methods *)
    let myunion m1 m2 = M.fold M.add m2 m1
    let mymap f m = M.fold (fun k e m -> M.add k (f k e) m) m M.empty
    let format start finish sep p m =
      let rec pr k e needSep =
	if needSep then
	  (Format.print_string sep; Format.print_cut(); p k e; true)
	else
	  (p k e; true)
    in
      Format.print_string start;
      Format.open_hovbox 0; let _ = M.fold pr m false in ();
      Format.print_string finish;
      Format.close_box()
  end


module StringMap =
  Make(struct
    type t = string
    let compare = compare
  end)

module IntMap =
  Make(struct
    type t = int
    let compare = (-)
  end)

let formatList start finish sep p l =
  let rec pr needSep = function
    [] -> ()
  | (h :: t) ->
    if needSep then
      (Format.print_string sep;
       Format.print_cut(); p h;
       pr true t)
    else
      (p h; pr true t)
in
  Format.print_string start;
  Format.open_hovbox 0;
  pr false l;
  Format.print_string finish;
  Format.close_box()

let listToString start finish sep p l =
  let rec pr needSep = function
      [] -> ""
    | (h :: t) ->
    	if needSep then
	  (sep^(p h)^(pr true t))
    	else
	  (p h^(pr true t))
  in
  start^(pr false l)^finish

let printList start finish sep p os l =
  let rec pr needSep = function
    [] -> ()
  | (h :: t) ->
    if needSep then
      (output_string os sep; p os h; pr true t)
    else
      (p os h; pr true t)
in
  output_string os start; pr false l; output_string os finish

let itFold f b l =
  let rec iterate i b = function
    [] -> b
  | (x::xs) -> iterate (succ i) (f i x b) xs
in
  iterate 0 b l
