
(* Integer operations *)

val iterateInt: int -> int -> int -> (int -> 'a -> 'a) -> 'a -> 'a

(* List operations *)

val itFold: (int -> 'a -> 'b -> 'b) -> 'b -> 'a list -> 'b
val formatList : string -> string -> string -> ('a -> unit) -> 'a list -> unit
val listToString : string -> string -> string -> ('a -> string) -> 'a list -> string
val printList :
  string -> string -> string -> (out_channel -> 'a -> unit) ->
  out_channel -> 'a list -> unit

(* Ready-built modules *)

module StringSet : Set.S with type elt = string

(* Extension of standard Map module *)

(* old, working with Ocaml-1.05 
module type S =
  sig
    type key
    type 'a t
    val empty: 'a t
    val add: key -> 'a -> 'a t -> 'a t
    val find: key -> 'a t -> 'a
    val remove: key -> 'a t -> 'a t
    val iter: (key -> 'a -> 'b) -> 'a t -> unit
    val fold: (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
    val union : 'a t -> 'a t -> 'a t
    val map : (key -> 'a -> 'b) -> 'a t -> 'b t
    val format :
      string -> string -> string -> (key -> 'a -> unit) -> 'a t -> unit
  end
*)

(* old, working with Ocaml-3.0 ?
module type S =
  sig
    type key
    type +'a t
    val empty: 'a t
    val add: key -> 'a -> 'a t -> 'a t
    val find: key -> 'a t -> 'a
    val remove: key -> 'a t -> 'a t
    val mem : key -> 'a t -> bool
    val iter : (key -> 'a -> unit) -> 'a t -> unit
    val fold: (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
    val map : ('a -> 'b) -> 'a t -> 'b t
(*
    val mapi : f:(key -> 'a -> 'b) -> 'a t -> 'b t
*)
    val mapi : (key -> 'a -> 'b) -> 'a t -> 'b t
    val union : 'a t -> 'a t -> 'a t
    val mymap : (key -> 'a -> 'b) -> 'a t -> 'b t
    val format :
      string -> string -> string -> (key -> 'a -> unit) -> 'a t -> unit
  end

*)

module type S =
  sig
    type key
    type +'a t
    val empty : 'a t
    val is_empty : 'a t -> bool
    val mem : key -> 'a t -> bool
    val add : key -> 'a -> 'a t -> 'a t
    val update : key -> ('a option -> 'a option) -> 'a t -> 'a t
    val singleton : key -> 'a -> 'a t
    val remove : key -> 'a t -> 'a t
    val merge : (key -> 'a option -> 'b option -> 'c option) -> 'a t -> 'b t -> 'c t
(*
  val union : 'a t -> 'a t -> 'a t  (* reimplemented *)
 *)
    val union : (key -> 'a -> 'a -> 'a option) -> 'a t -> 'a t -> 'a t
    val compare : ('a -> 'a -> int) -> 'a t -> 'a t -> int
    val equal : ('a -> 'a -> bool) -> 'a t -> 'a t -> bool
    val iter : (key -> 'a -> unit) -> 'a t -> unit
    val fold : (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
    val for_all : (key -> 'a -> bool) -> 'a t -> bool
    val exists : (key -> 'a -> bool) -> 'a t -> bool
    val filter : (key -> 'a -> bool) -> 'a t -> 'a t
    val partition : (key -> 'a -> bool) -> 'a t -> 'a t * 'a t
    val cardinal : 'a t -> int
    val bindings : 'a t -> (key * 'a) list
    val min_binding : 'a t -> key * 'a
    val min_binding_opt : 'a t -> (key * 'a) option
    val max_binding : 'a t -> key * 'a
    val max_binding_opt : 'a t -> (key * 'a) option
    val choose : 'a t -> key * 'a
    val choose_opt : 'a t -> (key * 'a) option
    val split : key -> 'a t -> 'a t * 'a option * 'a t	      
    val find : key -> 'a t -> 'a
    val find_opt : key -> 'a t -> 'a option
    val find_first : (key -> bool) -> 'a t -> key * 'a
    val find_first_opt : (key -> bool) -> 'a t -> (key * 'a) option
    val find_last : (key -> bool) -> 'a t -> key * 'a
    val find_last_opt : (key -> bool) -> 'a t -> (key * 'a) option
    val map : ('a -> 'b) -> 'a t -> 'b t
    val mapi : (key -> 'a -> 'b) -> 'a t -> 'b t
    val to_seq : 'a t -> (key * 'a) Stdlib.Seq.t
    val to_seq_from : key -> 'a t -> (key * 'a) Stdlib.Seq.t
    val add_seq : (key * 'a) Stdlib.Seq.t -> 'a t -> 'a t
    val of_seq : (key * 'a) Stdlib.Seq.t -> 'a t	
	(* added new methods *)
    val myunion : 'a t -> 'a t -> 'a t  (* reimplemented *)
    val mymap : (key -> 'a -> 'b) -> 'a t -> 'b t
    val format :
      string -> string -> string -> (key -> 'a -> unit) -> 'a t -> unit
  end


module Make (Ord:Map.OrderedType): (S with type key = Ord.t)

(*
 * Ready-build map implementations
 *)

module StringMap : S with type key = string
module IntMap : Map.S with type key = int

