(*
 * Applying the compositional translation.
 *)


(* TO DO:  

- now we generate lots of meaningless PRL and SKIP. Maybe replacing 
  PROC by DEC as continuation would help? (The optmiser seems doing 
  better job for declarations than for processes.); see also below

- agent a=P in Q requires Q to be a process; this doesn't work for
  def f(..):T = (agent a = P in a) and instead 'a' must be replaced 
  by process r!a where r:/T
  
*)


open Syntax

let thisAgentMsgType  = ref (Enco.TVAR(Error.unknown,Enco.fresh ""))
  
(*
 * Replacing the high-level constructs by their definitions.
 *)

let rec pluginProc ec pa e = function
    Enco.PRL(fi,p,q) -> PRL(fi,pluginProc ec pa e p,pluginProc ec pa e q)
  | Enco.IF(fi,v,p,q) -> IF(fi,pluginVal ec pa e v,pluginProc ec pa e p,pluginProc ec pa e q)
  | Enco.INPUT(fi,v,pat,p) -> INPUT(fi,pluginVal ec pa e v,pluginPat ec pa e pat,pluginProc ec pa e p)
  | Enco.OUTPUT(fi,v,w) -> OUTPUT(fi,pluginVal ec pa e v,pluginVal ec pa e w)

  | Enco.INPA(fi,v,pat,p) -> INPA(fi,pluginPat ec pa e pat,pluginProc ec pa e p)
  | Enco.OUTA(fi,v,w) -> OUTA(fi,pluginVal ec pa e v,pluginVal ec pa e w)
  | Enco.LDOUTA(fi,u,v,w) -> LDOUTA(fi,pluginVal ec pa e u,pluginVal ec pa e v,pluginVal ec pa e w)
  | Enco.TRYHERE(fi,v,w,p,q) -> TRYHERE(fi,pluginVal ec pa e v,pluginVal ec pa e w,pluginProc ec pa e p,pluginProc ec pa e q)

  | Enco.REP_INPUT(fi,v,pat,p) -> REP_INPUT(fi,pluginVal ec pa e v,pluginPat ec pa e pat,pluginProc ec pa e p)
  | Enco.TIMED_INPUT(fi,w,v,pat,p,q) -> TIMED_INPUT(fi,pluginVal ec pa e w,pluginVal ec pa e v,
							pluginPat ec pa e pat,pluginProc ec pa e p,pluginProc ec pa e q)
  | Enco.REP_TIMED_INPUT(fi,w,v,pat,p,q) -> REP_TIMED_INPUT(fi,pluginVal ec pa e w,pluginVal ec pa e v,
							pluginPat ec pa e pat,pluginProc ec pa e p,pluginProc ec pa e q)
  | Enco.LI_SEND(fi,v,u,w) -> LI_SEND(fi,pluginVal ec pa e v,pluginVal ec pa e u,pluginVal ec pa e w)
  | Enco.REMOTE_SEND(fi,v,s,u,w) -> REMOTE_SEND(fi,pluginVal ec pa e v,pluginVal ec pa e s,pluginVal ec pa e u,pluginVal ec pa e w)
  | Enco.LOCAL_SEND(fi,v,u,w) -> LOCAL_SEND(fi,pluginVal ec pa e v,pluginVal ec pa e u,pluginVal ec pa e w)
  | Enco.IFLOCAL(fi,v,u,w,p,q) -> IFLOCAL(fi,pluginVal ec pa e v,pluginVal ec pa e u,pluginVal ec pa e w,pluginProc ec pa e p,pluginProc ec pa e q)

  | Enco.LET(fi,d,p) -> LET(fi,pluginDec ec pa e d,pluginProc ec pa e p)
  | Enco.SKIP(fi) -> SKIP(fi)
  | Enco.TERMINATE(fi) -> TERMINATE(fi)
  | Enco.TYPECASE(fi,v,l,p) ->
      TYPECASE(fi,pluginVal ec pa e v,
	       List.map (fun (pat,p) -> (pluginPat ec pa e pat,pluginProc ec pa e p)) l,
	       pluginProc ec pa e p)
  | Enco.CASE(fi,topt,v,l) ->
      CASE(fi,pluginTy ec pa e topt,pluginVal ec pa e v,
	   List.map (fun (s,pat,p) -> (s,pluginPat ec pa e pat,pluginProc ec pa e p)) l)
  | Enco.ENCODE_PROC(fi,x,param) ->
      let p = Enco.getProc fi x e in
      encoProc ec (Enco.getId fi param e) e p

and pluginPat ec pa e = function
    Enco.RECORDp(fi,l) -> RECORDp(fi,List.map (pluginFieldPat ec pa e) l)
  | Enco.ASp(fi,x,topt,pat) -> ASp(fi,Enco.getName fi x e, pluginTy ec pa e topt,pluginPat ec pa e pat)
  | Enco.VARp(fi,x,topt) -> VARp(fi,Enco.getName fi x e,pluginTy ec pa e topt)
  | Enco.WILDp(fi,topt) -> WILDp(fi,pluginTy ec pa e topt)
  | Enco.RECp(fi,topt,pat) -> RECp(fi,pluginTy ec pa e topt,pluginPat ec pa e pat)
  | Enco.ENCODE_PAT(fi,x) ->
      let pat = Enco.getPat fi x e in
      encoPat ec pa e pat
	
and pluginFieldPat ec pa e = function
    Enco.ANONp(fi,pat) -> ANONp(fi,pluginPat ec pa e pat)
  | Enco.LABELEDp(fi,s,pat) -> LABELEDp(fi,s,pluginPat ec pa e pat)
  | Enco.TYp(fi,x,constr) -> TYp(fi,Enco.getName fi x e,pluginConstr ec pa e constr)

and pluginDec ec pa e = function
    Enco.SEQ(fi,d1,d2) ->
      let d1 = pluginDec ec pa e d1 in
      let d2 = pluginDec ec pa e d2 in
      SEQ(fi,d1,d2)
  | Enco.DEF(fi,l) ->
      (DEF(fi,List.map (fun (fi,n,pat,p) ->
      	(fi,Enco.getName fi n e,pluginPat ec pa e pat,pluginProc ec pa e p)) l))
  | Enco.INLINE(fi,x,pat,p) -> INLINE(fi,Enco.getName fi x e,pluginPat ec pa e pat,pluginProc ec pa e p)
  | Enco.VAL(fi,pat,v) -> VAL(fi,pluginPat ec pa e pat,pluginVal ec pa e v)
  | Enco.NEW(fi,x,t) -> NEW(fi,Enco.getName fi x e,pluginTy' ec pa e t)
  | Enco.RUN(fi,p) -> RUN(fi,pluginProc ec pa e p)
  | Enco.NEWLOC(fi,x) -> NEWLOC(fi,Enco.getName fi x e)
  | Enco.CREATE(fi,l) ->
      (CREATE(fi,List.map (fun (fi,n,t,p) ->
      	(fi,Enco.getName fi n e,pluginTy' ec pa e t,pluginProc ec pa e p)) l))
  | Enco.CREATE_AGENT(fi,l) ->
      (CREATE_AGENT(fi,List.map (fun (fi,n,p) ->
      	(fi,Enco.getName fi n e,pluginProc ec pa e p)) l))
  | Enco.AT(fi,x) -> AT(fi,Enco.getName fi x e)
  | Enco.MIGRATE(fi,v) -> MIGRATE(fi,pluginVal ec pa e v)
  | Enco.MACRO(fi,strkey,v) -> Enco.anotherPass := true; MACRO(fi,strkey,pluginVal ec pa e v)
  | Enco.MACROLET(fi,strkey,x,v) -> Enco.anotherPass := true; MACROLET(fi,strkey,Enco.getName fi x e,pluginVal ec pa e v)
  | Enco.INLINE_DEC(fi,dec) -> pluginDec ec pa e dec
  | Enco.INLINE_VAL(fi,v) -> pluginDec ec pa e (Enco.VAL(fi,Enco.VARp(fi,pa,None),v))
  | Enco.DEFER_DEC(fi,d) -> d
  | Enco.ENCODE_DEC(fi,x,param) ->
      let dec = Enco.getDec fi x e in
      encoDec ec (Enco.getId fi param e) e dec (Enco.SKIP(fi))

and pluginTy' ec pa e = function
    Enco.TOP(fi) -> TOP(fi)
  | Enco.TVAR(fi,id) -> TVAR(fi,Enco.name id)
  | Enco.TABS(fi,l,t) -> TABS(fi,List.map (fun (id,k) -> (Enco.name id, k)) l, pluginTy' ec pa e t)
  | Enco.REC(fi,id,k,t) -> REC(fi,Enco.name id,k,pluginTy' ec pa e t)

  | Enco.TAPP(fi,t,l) -> TAPP(fi,pluginTy' ec pa e t,List.map (fun t -> pluginTy' ec pa e t) l)
  | Enco.RECORD(fi,l) -> RECORD(fi,List.map (fun ft -> pluginFieldTy ec pa e ft) l)
  | Enco.WITH(fi,t,l) -> WITH(fi,pluginTy' ec pa e t,List.map (fun ft -> pluginFieldTy ec pa e ft) l)
  | Enco.WHERE(fi,t,l) -> WHERE(fi,pluginTy' ec pa e t,List.map (fun ft -> pluginFieldTy ec pa e ft) l)
  | Enco.CHAN(fi,t) -> CHAN(fi,pluginTy' ec pa e t)
  | Enco.OCHAN(fi,t) -> OCHAN(fi,pluginTy' ec pa e t)
  | Enco.ICHAN(fi,t) -> ICHAN(fi,pluginTy' ec pa e t)
  | Enco.RCHAN(fi,t) -> RCHAN(fi,pluginTy' ec pa e t)
  | Enco.STRING(fi) -> STRING(fi)
  | Enco.BOOL(fi) -> BOOL(fi)
  | Enco.CHAR(fi) -> CHAR(fi)
  | Enco.INT(fi) -> INT(fi)
  | Enco.SITE(fi) -> SITE(fi)
  | Enco.LOC(fi) -> LOC(fi)
  | Enco.AGENTT(fi,t) -> AGENTT(fi,pluginTy' ec pa e t)
  | Enco.AGENT(fi) -> AGENT(fi)
  | Enco.DYN(fi) -> DYN(fi)
  | Enco.VART(fi,l) -> VART(fi,List.map (fun (s,t) -> (s,pluginTy' ec pa e t)) l)
  | Enco.MAP(fi,t1,t2) -> MAP(fi,pluginTy' ec pa e t1,pluginTy' ec pa e t2)
  | Enco.ENCODE_TYPE(fi,x) ->
      let t = Enco.getTy fi x e in
      encoTy' ec pa e t

and pluginFieldTy ec pa e = function
    Enco.ANON(fi,t) -> ANON(fi,pluginTy' ec pa e t)
  | Enco.LABELED(fi,s,t) -> LABELED(fi,s,pluginTy' ec pa e t)
  | Enco.TYP(fi,id,constr) -> TY(fi,Enco.name id,pluginConstr ec pa e constr)

and pluginTy ec pa e = function
    None -> None
  | Some(t) -> Some(pluginTy' ec pa e t)
	
and pluginConstr ec pa e = function
    Enco.LT(t) -> LT(pluginTy' ec pa e t)
  | Enco.EQ(t) -> EQ(pluginTy' ec pa e t)

and pluginVal ec pa e = function
    Enco.CCODEv(fi,ci,l) ->
      CCODEv(fi,pluginCcode e ci,List.map (fun fv -> pluginFieldVal ec pa e fv) l)
  | Enco.APPv(fi,topt,v,l) ->
      APPv(fi,pluginTy ec pa e topt,pluginVal ec pa e v,List.map (fun fv -> pluginFieldVal ec pa e fv) l)
  | Enco.RECORDv(fi,l) -> RECORDv(fi,List.map (fun fv -> pluginFieldVal ec pa e fv) l)
  | Enco.WITHv(fi,topt,v,l) ->
      WITHv(fi,pluginTy ec pa e topt,pluginVal ec pa e v,List.map (fun fv -> pluginFieldVal ec pa e fv) l)
  | Enco.WHEREv(fi,topt,v,l) ->
      WHEREv(fi,pluginTy ec pa e topt,pluginVal ec pa e v,List.map (fun fv -> pluginFieldVal ec pa e fv) l)
  | Enco.STRINGv(fi,s) -> STRINGv(fi,s)
  | Enco.SITEv(fi,s,i) -> SITEv(fi,s,i)
  | Enco.SELECTv(fi,v,s) -> SELECTv(fi,pluginVal ec pa e v,s)
  | Enco.VARv(fi,x) -> VARv(fi,Enco.getName fi x e)
  | Enco.BOOLv(fi,b) -> BOOLv(fi,b)
  | Enco.CHARv(fi,c) -> CHARv(fi,c)
  | Enco.INTv(fi,i) -> INTv(fi,i)
  | Enco.LETv(fi,d,v) -> LETv(fi,pluginDec ec pa e d,pluginVal ec pa e v)
  | Enco.RECv(fi,v,topt) -> RECv(fi,pluginVal ec pa e v,pluginTy ec pa e topt)
  | Enco.IFv(fi,topt,u,v,w) -> IFv(fi,pluginTy ec pa e topt,pluginVal ec pa e u,pluginVal ec pa e v,pluginVal ec pa e w)
  | Enco.ABSv(fi,pat,p) -> ABSv(fi,pluginPat ec pa e pat,pluginProc ec pa e p)
  | Enco.DYNAMICv(fi,v,topt) -> DYNAMICv(fi,pluginVal ec pa e v,pluginTy ec pa e topt)
  | Enco.TYPECASEv(fi,topt,v,l,w) ->
      TYPECASEv(fi,pluginTy ec pa e topt,pluginVal ec pa e v,
		List.map (fun (pat,u) -> (pluginPat ec pa e pat,pluginVal ec pa e u)) l,
		pluginVal ec pa e w)
  | Enco.VARIANTv(fi,s,v) -> VARIANTv(fi,s,pluginVal ec pa e v)
  | Enco.CASEv(fi,topt,v,l) ->
      CASEv(fi,pluginTy ec pa e topt,pluginVal ec pa e v,
	    List.map (fun (s,pat,u) -> (s,pluginPat ec pa e pat,pluginVal ec pa e u)) l)
  | Enco.ENCODE_VALUE(fi,x) ->
      let v = Enco.getVal fi x e in
      encoVal ec pa e v

and pluginFieldVal ec pa e = function
    Enco.ANONv(fi,v) -> ANONv(fi,pluginVal ec pa e v)
  | Enco.LABELEDv(fi,s,v) -> LABELEDv(fi,s,pluginVal ec pa e v)
  | Enco.TYv(fi,t) -> TYv(fi,pluginTy' ec pa e t)
      	
and pluginCcode e ci = ci


(*
 * Applying definitions of high-level constructs.
 *)

and encoPat ec pa e = function
    Enco.RECORDp(fi,l) -> RECORDp(fi,List.map (encoFieldPat ec pa e) l)
  | Enco.ASp(fi,n,topt,pat) -> ASp(fi,Enco.name n,encoTy ec pa e topt,encoPat ec pa e pat)
  | Enco.VARp(fi,n,topt) -> VARp(fi,Enco.name n,encoTy ec pa e topt)
  | Enco.WILDp(fi,topt) -> WILDp(fi,encoTy ec pa e topt)
  | Enco.RECp(fi,topt,pat) -> RECp(fi,encoTy ec pa e topt,encoPat ec pa e pat)
  | Enco.ENCODE_PAT(fi,_) -> Error.errAt fi "This pattern is allowed in translation definitions only." 
	
and encoFieldPat ec pa e = function
  | Enco.ANONp(fi,pat) -> ANONp(fi,encoPat ec pa e pat)
  | Enco.LABELEDp(fi,label,pat) -> LABELEDp(fi,label,encoPat ec pa e pat)
  | Enco.TYp(fi,n,constr) -> TYp(fi,Enco.name n,encoConstr ec pa e constr)

(*
(* encode pattern with type translation for typecase *)
and encoPatPluginTy ec pa e = function
    Enco.RECORDp(fi,l) -> RECORDp(fi,List.map (encoFieldPatPluginTy ec pa e) l)
  | Enco.ASp(fi,n,topt,pat) -> ASp(fi,Enco.name n,pluginTy ec pa e topt,encoPatPluginTy ec pa e pat)
  | Enco.VARp(fi,n,topt) -> VARp(fi,Enco.name n,pluginTy ec pa e topt)
  | Enco.WILDp(fi,topt) -> WILDp(fi,pluginTy ec pa e topt)
  | Enco.RECp(fi,topt,pat) -> RECp(fi,pluginTy ec pa e topt,encoPatPluginTy ec pa e pat)
  | Enco.ENCODE_PAT(fi,_) -> Error.errAt fi "This pattern is allowed in translation definitions only." 
	
and encoFieldPatPluginTy ec pa e = function
  | Enco.ANONp(fi,pat) -> ANONp(fi,encoPatPluginTy ec pa e pat)
  | Enco.LABELEDp(fi,label,pat) -> LABELEDp(fi,label,encoPatPluginTy ec pa e pat)
  | Enco.TYp(fi,n,constr) -> TYp(fi,Enco.name n,encoConstrPluginTy ec pa e constr)
*)

	
and encoProc ec pa e = function
    Enco.PRL(fi,p,q) ->
      let p = encoProc ec pa e p in
      let q = encoProc ec pa e q in
      PRL(fi,p,q)
  | Enco.IF(fi,v,p,q) ->
      (try match Misc.StringMap.find "IF" ec.Enco.forms with
      	Enco.DEF_PROC(l,def) -> 
	  let (v,def) = Enco.modifyProc fi v def in
	  pluginProc ec pa
            (Enco.bind fi l (Enco.NAME(v)::Enco.PROC(p)::Enco.PROC(q)
			     ::Enco.NAME(pa)::[]) Enco.enull) def
      | _ -> raise Not_found
      with
      	Not_found -> IF(fi,encoVal ec pa e v,encoProc ec pa e p,
              		encoProc ec pa e q))
  | Enco.INPUT(fi,v,pat,p) ->
      (try match Misc.StringMap.find "INPUT" ec.Enco.forms with
      	Enco.DEF_PROC(l,def) -> 
	  let (v,def) = Enco.modifyProc fi v def in
	  pluginProc ec pa
            (Enco.bind fi l (Enco.NAME(v)::Enco.PAT(pat)::Enco.PROC(p)
			     ::Enco.NAME(pa)::[]) Enco.enull) def
      | _ -> raise Not_found
      with
      	Not_found -> INPUT(fi,encoVal ec pa e v,encoPat ec pa e pat,
			   encoProc ec pa e p))
  | Enco.OUTPUT(fi,v,w) ->
      (try match Misc.StringMap.find "OUTPUT" ec.Enco.forms with
      	Enco.DEF_PROC(l,def) -> 
	  let (w,def) = Enco.modifyProc fi w def in
	  let (v,def) = Enco.modifyProc fi v def in
	  pluginProc ec pa
            (Enco.bind fi l (Enco.NAME(v)::Enco.NAME(w)
			     ::Enco.NAME(pa)::[]) Enco.enull) def
      | _ -> raise Not_found
      with
      	Not_found -> OUTPUT(fi,encoVal ec pa e v,encoVal ec pa e w))
  | Enco.INPA(fi,_,pat,p) ->
      (try match Misc.StringMap.find "INPA" ec.Enco.forms with
	Enco.DEF_PROC(l,def) -> pluginProc ec pa
            (Enco.bind fi l (Enco.PAT(pat)::Enco.PROC(p)
			     ::Enco.NAME(pa)::[]) Enco.enull) def
      | _ -> raise Not_found
      with
	Not_found -> INPA(fi,encoPat ec pa e pat,encoProc ec pa e p))
  | Enco.OUTA(fi,v,w) ->
      (try match Misc.StringMap.find "OUTA" ec.Enco.forms with
      	Enco.DEF_PROC(l,def) -> pluginProc ec pa
            (Enco.bind fi l (Enco.VALUE(v)::Enco.VALUE(w)
			     ::Enco.NAME(pa)::[]) Enco.enull) def
      | _ -> raise Not_found
      with
      	Not_found -> OUTA(fi,encoVal ec pa e v,encoVal ec pa e w))
  | Enco.LDOUTA(fi,u,v,w) ->
      (try match Misc.StringMap.find "LDOUTA" ec.Enco.forms with
      	Enco.DEF_PROC(l,def) -> pluginProc ec pa
            (Enco.bind fi l (Enco.VALUE(u)::(Enco.TY(!thisAgentMsgType))::Enco.VALUE(v)::Enco.VALUE(w)
			     ::Enco.NAME(pa)::[]) Enco.enull) def
      | _ -> raise Not_found
      with
      	Not_found -> LDOUTA(fi,encoVal ec pa e u,encoVal ec pa e v,
			    encoVal ec pa e w))
  | Enco.TRYHERE(fi,v,w,p,q) ->
      (try match Misc.StringMap.find "TRYHERE" ec.Enco.forms with
      	Enco.DEF_PROC(l,def) -> pluginProc ec pa
            (Enco.bind fi l (Enco.VALUE(v)::Enco.TY(!thisAgentMsgType)::Enco.VALUE(w)::Enco.PROC(p)
			     ::Enco.PROC(q)::Enco.NAME(pa)::[]) Enco.enull) def
      | _ -> raise Not_found
      with
      	Not_found -> TRYHERE(fi,encoVal ec pa e v,encoVal ec pa e w,
			     encoProc ec pa e p,encoProc ec pa e q))
  | Enco.REP_INPUT(fi,v,pat,p) ->
      (try match Misc.StringMap.find "REP_INPUT" ec.Enco.forms with
      	Enco.DEF_PROC(l,def) -> 
	  let (v,def) = Enco.modifyProc fi v def in
	  pluginProc ec pa
            (Enco.bind fi l (Enco.NAME(v)::Enco.PAT(pat)::Enco.PROC(p)
			     ::Enco.NAME(pa)::[]) Enco.enull) def
      | _ -> raise Not_found
      with
      	Not_found -> REP_INPUT(fi,encoVal ec pa e v,encoPat ec pa e pat,
			       encoProc ec pa e p))
  | Enco.TIMED_INPUT(fi,w,v,pat,p,q) ->
      (try match Misc.StringMap.find "TIMED_INPUT" ec.Enco.forms with
	Enco.DEF_PROC(l,def) ->
	  let (v,def) = Enco.modifyProc fi v def in
	  let (w,def) = Enco.modifyProc fi w def in
 	  pluginProc ec pa
            (Enco.bind fi l (Enco.NAME(w)::Enco.NAME(v)::Enco.PAT(pat)::Enco.PROC(p)::Enco.PROC(q)
			     ::Enco.NAME(pa)::[]) Enco.enull) def
      | _ -> raise Not_found
      with
	Not_found -> TIMED_INPUT(fi,encoVal ec pa e w,encoVal ec pa e v,encoPat ec pa e pat,
				 encoProc ec pa e p,encoProc ec pa e q))
  | Enco.REP_TIMED_INPUT(fi,w,v,pat,p,q) ->
      (try match Misc.StringMap.find "REP_TIMED_INPUT" ec.Enco.forms with
	Enco.DEF_PROC(l,def) -> 
	  let (v,def) = Enco.modifyProc fi v def in
	  let (w,def) = Enco.modifyProc fi w def in
	  pluginProc ec pa
            (Enco.bind fi l (Enco.NAME(w)::Enco.NAME(v)::Enco.PAT(pat)::Enco.PROC(p)::Enco.PROC(q)
			     ::Enco.NAME(pa)::[]) Enco.enull) def
      | _ -> raise Not_found
      with
	Not_found -> REP_TIMED_INPUT(fi,encoVal ec pa e w,encoVal ec pa e v,encoPat ec pa e pat,
				     encoProc ec pa e p,encoProc ec pa e q))
  | Enco.LI_SEND(fi,v,u,w) ->
      (try match Misc.StringMap.find "LI_SEND" ec.Enco.forms with
      	Enco.DEF_PROC(l,def) -> 
          let (w,def) = Enco.modifyProc fi w def in
	  let (u,def) = Enco.modifyProc fi u def in
	  let (v,def) = Enco.modifyProc fi v def in
	  pluginProc ec pa
            (Enco.bind fi l (Enco.NAME(v)::Enco.NAME(u)::Enco.NAME(w)
			     ::Enco.NAME(pa)::[]) Enco.enull) def
      | _ -> raise Not_found
      with
      	Not_found -> LI_SEND(fi,encoVal ec pa e v,encoVal ec pa e u,encoVal ec pa e w))
  | Enco.REMOTE_SEND(fi,v,s,u,w) ->
      (try match Misc.StringMap.find "REMOTE_SEND" ec.Enco.forms with
      	Enco.DEF_PROC(l,def) -> 
          let (w,def) = Enco.modifyProc fi w def in
          let (u,def) = Enco.modifyProc fi u def in
          let (s,def) = Enco.modifyProc fi s def in
	  let (v,def) = Enco.modifyProc fi v def in 
	  pluginProc ec pa
            (Enco.bind fi l (Enco.NAME(v)::Enco.NAME(s)::Enco.NAME(u)::Enco.NAME(w)
			     ::Enco.NAME(pa)::[]) Enco.enull) def
      | _ -> raise Not_found
      with
      	Not_found -> REMOTE_SEND(fi,encoVal ec pa e v,encoVal ec pa e s,encoVal ec pa e u,encoVal ec pa e w))
  | Enco.LOCAL_SEND(fi,v,u,w) ->
      (try match Misc.StringMap.find "LOCAL_SEND" ec.Enco.forms with
      	Enco.DEF_PROC(l,def) -> 
          let (w,def) = Enco.modifyProc fi w def in
          let (u,def) = Enco.modifyProc fi u def in
	  let (v,def) = Enco.modifyProc fi v def in 
	  pluginProc ec pa
            (Enco.bind fi l (Enco.NAME(v)::Enco.NAME(u)::Enco.NAME(w)
			     ::Enco.NAME(pa)::[]) Enco.enull) def
      | _ -> raise Not_found
      with
      	Not_found -> LOCAL_SEND(fi,encoVal ec pa e v,encoVal ec pa e u,encoVal ec pa e w))
  | Enco.IFLOCAL(fi,v,u,w,p,q) ->
      (try match Misc.StringMap.find "IFLOCAL" ec.Enco.forms with
      	Enco.DEF_PROC(l,def) -> 
          let (w,def) = Enco.modifyProc fi w def in
          let (u,def) = Enco.modifyProc fi u def in
	  let (v,def) = Enco.modifyProc fi v def in 
	  pluginProc ec pa
            (Enco.bind fi l (Enco.NAME(v)::Enco.NAME(u)::Enco.NAME(w)::Enco.PROC(p)
			     ::Enco.PROC(q)::Enco.NAME(pa)::[]) Enco.enull) def
      | _ -> raise Not_found
      with
      	Not_found -> IFLOCAL(fi,encoVal ec pa e v,encoVal ec pa e u,encoVal ec pa e w,
			     encoProc ec pa e p,encoProc ec pa e q))
  | Enco.LET(fi,d,p) ->
      let d = encoDec ec pa e d p in
      LET(fi,d,SKIP(fi))
  | Enco.SKIP(fi) ->
      (try match Misc.StringMap.find "SKIP" ec.Enco.forms with
      	Enco.DEF_PROC(l,def) -> pluginProc ec pa
            (Enco.bind fi l (Enco.NAME(pa)::[]) Enco.enull) def
      | _ -> raise Not_found
      with
      	Not_found -> SKIP(fi))
  | Enco.TERMINATE(fi) ->
      (try match Misc.StringMap.find "TERMINATE" ec.Enco.forms with
      	Enco.DEF_PROC(l,def) -> pluginProc ec pa
            (Enco.bind fi l (Enco.NAME(pa)::[]) Enco.enull) def
      | _ -> raise Not_found
      with
      	Not_found -> TERMINATE(fi))
  | Enco.TYPECASE(fi,v,cl,q) ->
      (try match Misc.StringMap.find "TYPECASE" ec.Enco.forms with
      	Enco.DEF_PROC(l,def) -> 
	  let (v,def) = Enco.modifyProc fi v def in 
	  pluginProc ec pa
            (Enco.bind fi l (Enco.NAME(v)
			     ::Enco.LIST(List.map (fun (pat,p) -> Enco.TUPLE(Enco.PAT(pat),Enco.PROC(p))) cl)
			     ::Enco.PROC(q)
			     ::Enco.NAME(pa)::[]) Enco.enull) def
      | _ -> raise Not_found
      with

      	Not_found -> TYPECASE(fi,encoVal ec pa e v,
			      List.map (fun (pat,p) -> 
				(encoPat ec pa e pat,encoProc ec pa e p)) cl,
			      encoProc ec pa e q))
  | Enco.CASE(fi,topt,v,cl) ->
      (try match Misc.StringMap.find "CASE" ec.Enco.forms with
	Enco.DEF_PROC(l,def) -> 
          let (v,def) = Enco.modifyProc fi v def in
	  pluginProc ec pa
            (Enco.bind fi l (Enco.TYOPT(topt)::Enco.NAME(v)
			     ::Enco.LIST(List.map (fun (s,pat,p) -> Enco.TRIPLE(Enco.VALUE(Enco.STRINGv(fi,s)),
										Enco.PAT(pat),Enco.PROC(p))) cl)
			     ::Enco.NAME(pa)::[]) Enco.enull) def
      | _ -> raise Not_found
      with
	Not_found -> CASE(fi,encoTy ec pa e topt, encoVal ec pa e v,
          		  List.map (fun (s,pat,p) -> (s,encoPat ec pa e pat, encoProc ec pa e p)) 
			    cl))
  | Enco.ENCODE_PROC(fi,_,_) -> Error.errAt fi "This construct is allowed in translation definitions only."

and encoVal ec pa e = function
    Enco.CCODEv(fi,ci,l) -> CCODEv(fi,ci,List.map (encoFieldVal ec pa e) l)
  | Enco.RECORDv(fi,l) -> RECORDv(fi,List.map (encoFieldVal ec pa e) l)
  | Enco.WITHv(fi,topt,v,l) -> WITHv(fi,encoTy ec pa e topt,encoVal ec pa e v,
				     List.map (encoFieldVal ec pa e) l)
  | Enco.WHEREv(fi,topt,v,l) -> WHEREv(fi,encoTy ec pa e topt,encoVal ec pa e v,
				       List.map (encoFieldVal ec pa e) l)
  | Enco.APPv(fi,topt,v,l) -> APPv(fi,encoTy ec pa e topt,encoVal ec pa e v,
				   List.map (encoFieldVal ec pa e) l)
  | Enco.SELECTv(fi,v,s) -> SELECTv(fi,encoVal ec pa e v,s)
  | Enco.STRINGv(fi,s) -> STRINGv(fi,s)
  | Enco.SITEv(fi,s,i) -> SITEv(fi,s,i)
  | Enco.CHARv(fi,c) -> CHARv(fi,c)
  | Enco.VARv(fi,x) ->  VARv(fi,Enco.name x)
  | Enco.BOOLv(fi,b) -> BOOLv(fi,b)
  | Enco.INTv(fi,i) -> INTv(fi,i)
  | Enco.LETv(fi,d,v) -> LETv(fi,encoDec ec pa e d (Enco.SKIP(fi)),encoVal ec pa e v)
  | Enco.RECv(fi,v,topt) -> RECv(fi,encoVal ec pa e v,encoTy ec pa e topt)
  | Enco.IFv(fi,topt,u,v,w) -> IFv(fi,encoTy ec pa e topt,
				   encoVal ec pa e u,encoVal ec pa e v,encoVal ec pa e w)
  | Enco.ABSv(fi,pat,p) -> ABSv(fi,encoPat ec pa e pat,encoProc ec pa e p)
  | Enco.DYNAMICv(fi,v,topt) -> DYNAMICv(fi,encoVal ec pa e v,encoTy ec pa e topt)
  | Enco.TYPECASEv(fi,topt,v,l,velse) -> 
      TYPECASEv(fi,encoTy ec pa e topt,encoVal ec pa e v,
		List.map (fun (pat,u) -> 
		  (encoPat ec pa e pat,encoVal ec pa e u)) l,encoVal ec pa e velse)
  | Enco.VARIANTv(fi,label,v) -> VARIANTv(fi,label,encoVal ec pa e v)
  | Enco.CASEv(fi,topt,v,l) -> CASEv(fi,encoTy ec pa e topt,encoVal ec pa e v,
				     List.map (fun (label,pat,u) -> 
				       (label,encoPat ec pa e pat,encoVal ec pa e u)) l)
  | Enco.ENCODE_VALUE(fi,_) -> Error.errAt fi "This construct is allowed in translation definitions only."

	
and encoFieldVal ec pa e = function
    Enco.ANONv(fi,v) -> ANONv(fi,encoVal ec pa e v)
  | Enco.LABELEDv(fi,s,v) -> LABELEDv(fi,s,encoVal ec pa e v)
  | Enco.TYv(fi,t) -> TYv(fi,encoTy' ec pa e t)

	
and encoDec ec pa e dec cont = 
  match dec with
    Enco.SEQ(fi,d1,d2) ->
      let d = encoDec ec pa e d1 (Enco.LET(fi,d2,cont)) in d
  | Enco.DEF(fi,l) ->
      (match l with
      	[] -> retDec fi ec pa e (DEF(fi, [])) cont  (* not simply RUN(fi,encoProc ec pa e cont) ? *) 
      | ((infi,n,pat,p)::ld) ->
	  try match Misc.StringMap.find (Enco.str_of n) ec.Enco.funs with

(*
	    Enco.DEF_DEF(infi',n',pat',p') -> retDec fi ec pa e 
		(DEF(fi, [fi, Enco.name n,
			  pluginPat ec pa (Enco.bind fi (n'::[]) (Enco.NAME(n)::[]) Enco.enull) pat',
			  pluginProc ec pa (Enco.bind fi (n'::[]) (Enco.NAME(n)::[]) Enco.enull) p']))
		cont
*)
	    Enco.DEF_DEF(infi',n',pa',pat',p') -> retDec fi ec pa e 
		(DEF(fi, [fi, Enco.name n,
			  pluginPat ec pa (Enco.bind fi (n'::pa'::[]) (Enco.NAME(n)::Enco.NAME(pa)::[]) Enco.enull) pat',
			  pluginProc ec pa (Enco.bind fi (n'::pa'::[]) (Enco.NAME(n)::Enco.NAME(pa)::[]) Enco.enull) p']))
		cont

	  | _ -> raise Not_found
	  with
	    Not_found -> 
              (try match Misc.StringMap.find "DEF" ec.Enco.forms with
              	Enco.DEF_DEC(l,def) -> pluginDec ec pa 
		    (Enco.bind fi l (Enco.NAME(n)::Enco.PAT(pat)::Enco.PROC(p)
				     ::Enco.NAME(pa)::Enco.PROC(Enco.LET(fi,Enco.DEF(fi,ld),cont))::[])
		       Enco.enull) def
              | _ -> raise Not_found
              with
              	Not_found -> retDec fi ec pa e
		    (DEF(fi,List.map (fun (infi,n,pat,p) -> 
              	      (infi,Enco.name n,encoPat ec pa e pat,encoProc ec pa e p)) l))
		    cont))
  | Enco.INLINE(fi,n,pat,p) ->
      (try match Misc.StringMap.find "INLINE" ec.Enco.forms with
      	Enco.DEF_DEC(l,def) -> pluginDec ec pa
            (Enco.bind fi l (Enco.NAME(n)::Enco.PAT(pat)::Enco.PROC(p)
			     ::Enco.NAME(pa)::Enco.PROC(cont)::[]) Enco.enull) def
      | _ -> raise Not_found
      with
      	Not_found -> retDec fi ec pa e
            (INLINE(fi,Enco.name n,encoPat ec pa e pat,encoProc ec pa e p))
            cont)
  | Enco.VAL(fi,pat,v) ->
      (try match Misc.StringMap.find "VAL" ec.Enco.forms with
      	Enco.DEF_DEC(l,def) -> 
	  let (v,def) = Enco.modifyDec fi v def in
	  pluginDec ec pa
            (Enco.bind fi l (Enco.PAT(pat)::Enco.NAME(v)
			     ::Enco.NAME(pa)::Enco.PROC(cont)::[]) Enco.enull) def
      | _ -> raise Not_found
      with
      	Not_found -> retDec fi ec pa e
            (VAL(fi,encoPat ec pa e pat,encoVal ec pa e v)) cont)
  | Enco.NEW(fi,n,t) ->
      (try match Misc.StringMap.find "NEW" ec.Enco.forms with
      	Enco.DEF_DEC(l,def) -> pluginDec ec pa
            (Enco.bind fi l (Enco.NAME(n)::Enco.TY(t)
			     ::Enco.NAME(pa)::Enco.PROC(cont)::[]) Enco.enull) def
      | _ -> raise Not_found
      with
      	Not_found -> retDec fi ec pa e
            (NEW(fi,Enco.name n,encoTy' ec pa e t)) cont)
  | Enco.RUN(fi,p) ->
      (try match Misc.StringMap.find "RUN" ec.Enco.forms with
      	Enco.DEF_DEC(l,def) -> pluginDec ec pa
            (Enco.bind fi l (Enco.PROC(p)
			     ::Enco.NAME(pa)::Enco.PROC(cont)::[]) Enco.enull) def
      | _ -> raise Not_found
      with
      	Not_found -> retDec fi ec pa e
            (RUN(fi,encoProc ec pa e p)) cont)
  | Enco.NEWLOC(fi,n) ->
      (try match Misc.StringMap.find "NEWLOC" ec.Enco.forms with
      	Enco.DEF_DEC(l,def) -> pluginDec ec pa
            (Enco.bind fi l (Enco.NAME(n)
			     ::Enco.NAME(pa)::Enco.PROC(cont)::[]) Enco.enull) def
      | _ -> raise Not_found
      with
      	Not_found -> retDec fi ec pa e 
	    (NEWLOC(fi,Enco.name n)) cont)
  | Enco.CREATE(fi,l) ->
      (match l with
      	[] -> retDec fi ec pa e (DEF(fi, [])) cont  (* not CREATE(fi, []) ? *)
      | ((infi,n,t,p)::ld) ->
          (thisAgentMsgType := 
	    (match t with
	      Enco.AGENTT(_,x) -> x
	    | _ -> Error.bugAt infi "Plugin.encoDec: Enco.CREATE: Expected type Agent(..)");
 	   try match Misc.StringMap.find "CREATE" ec.Enco.forms with
             Enco.DEF_DEC(l,def) -> pluginDec ec pa
		 (Enco.bind fi l (Enco.NAME(n)::Enco.TY(!thisAgentMsgType)::Enco.PROC(p)
			       	  ::Enco.NAME(pa)::Enco.PROC(Enco.LET(fi,Enco.CREATE(fi,ld),cont))::[]) Enco.enull) def
     	   | _ -> raise Not_found
   	   with
             Not_found -> retDec fi ec pa e
		 (CREATE(fi,List.map (fun (infi,n,t,p) ->
              	   (infi,Enco.name n,encoTy' ec pa e t,encoProc ec pa e p)) l))
		 cont))
  | Enco.CREATE_AGENT(fi,l) ->
      (match l with
      	[] -> retDec fi ec pa e (DEF(fi, [])) cont
      | ((infi,n,p)::ld) ->
 	  try match Misc.StringMap.find "CREATE" ec.Enco.forms with
            Enco.DEF_DEC(l,def) -> pluginDec ec pa
              	(Enco.bind fi l (Enco.NAME(n)::Enco.PROC(p)
				 ::Enco.NAME(pa)::Enco.PROC(Enco.LET(fi,Enco.CREATE_AGENT(fi,ld),cont))::[]) Enco.enull) def
     	  | _ -> raise Not_found
   	  with
            Not_found -> retDec fi ec pa e
              	(CREATE_AGENT(fi,List.map (fun (infi,n,p) ->
              	  (infi,Enco.name n,encoProc ec pa e p)) l))
              	cont)
  | Enco.AT(fi,n) ->
      (try match Misc.StringMap.find "AT" ec.Enco.forms with
      	Enco.DEF_DEC(l,def) -> pluginDec ec pa
            (Enco.bind fi l (Enco.NAME(n)
			     ::Enco.NAME(pa)::Enco.PROC(cont)::[]) Enco.enull) def
      | _ -> raise Not_found
      with
      	Not_found -> retDec fi ec pa e
            (AT(fi,Enco.name n)) cont)
  | Enco.MIGRATE(fi,v) ->
      (try match Misc.StringMap.find "MIGRATE" ec.Enco.forms with
      	Enco.DEF_DEC(l,def) -> 
	  let (v,def) = Enco.modifyDec fi v def in
	  pluginDec ec pa
            (Enco.bind fi l (Enco.NAME(v)
			     ::Enco.NAME(pa)::Enco.PROC(cont)::[]) Enco.enull) def
      | _ -> raise Not_found
      with
      	Not_found -> retDec fi ec pa e
            (MIGRATE(fi,encoVal ec pa e v)) cont)
  | Enco.MACRO(fi,key,v) ->  (* plugin the macro definition *)
      (try match Misc.StringMap.find key ec.Enco.forms with
      	Enco.DEF_DEC(l,def) -> 
          let (v,def) = Enco.modifyDec fi v def in
	  pluginDec ec pa
            (Enco.bind fi l (Enco.NAME(v)	
			     ::Enco.NAME(pa)::Enco.PROC(cont)::[]) Enco.enull) def
      | _ -> raise Not_found
      with
      	Not_found -> Error.errAt fi ("Macro declaration "^key^" not found."))
(* 
 * These constructs below are currently not supported: 
 *)
  | Enco.MACROLET(fi,key,n,v) ->  (* plugin the macro definition "let..in.." *)
      (try match Misc.StringMap.find key ec.Enco.forms with
      	Enco.DEF_DEC(l,def) -> 
          let (v,def) = Enco.modifyDec fi v def in
	  pluginDec ec pa
            (Enco.bind fi l (Enco.NAME(n)::Enco.NAME(v)
			     ::Enco.NAME(pa)::Enco.PROC(cont)::[]) Enco.enull) def
      | _ -> raise Not_found
      with
      	Not_found -> Error.errAt fi ("Macro declaration let..in "^key^" not found."))
  | Enco.DEFER_DEC(fi,d) -> 
      let x = Enco.fresh "$CONTINUATION" in
      pluginDec ec pa (Enco.bind fi (x::[]) (Enco.PROC(cont)::[]) Enco.enull) 
      	(Enco.SEQ(fi,Enco.DEFER_DEC(fi,d),
		  Enco.ENCODE_DEC(fi,x,pa)))
  | Enco.INLINE_DEC(fi,dec) ->
      let x = Enco.fresh "$CONTINUATION" in
      pluginDec ec pa (Enco.bind fi (x::[]) (Enco.PROC(cont)::[]) Enco.enull) 
      	(Enco.SEQ(fi,dec,Enco.ENCODE_DEC(fi,x,pa)))
  | Enco.INLINE_VAL(fi,v) -> 
      let x = Enco.fresh "$CONTINUATION" in
      pluginDec ec pa (Enco.bind fi (x::[]) (Enco.PROC(cont)::[]) Enco.enull)
      	(Enco.SEQ(fi,Enco.VAL(fi,Enco.VARp(fi,pa,None),v),
		  Enco.ENCODE_DEC(fi,x,pa)))
(* end *)
  | Enco.ENCODE_DEC(fi,_,_) -> Error.errAt fi "This construct is allowed in translation definitions only." 

and encoTy' ec pa e = function
    Enco.TOP(fi) -> TOP(fi)
  | Enco.TVAR(fi,id) -> TVAR(fi,Enco.name id)
  | Enco.TABS(fi,l,t) -> TABS(fi,List.map (fun (id,k) -> 
      (Enco.name id,k)) l,encoTy' ec pa e t)
  | Enco.REC(fi,id,k,t) -> REC(fi,Enco.name id,k,encoTy' ec pa e t)
  | Enco.TAPP(fi,t,l) -> TAPP(fi,encoTy' ec pa e t,List.map (fun t -> encoTy' ec pa e t) l)
  | Enco.RECORD(fi,l) -> RECORD(fi,List.map (fun ft -> encoFieldTy ec pa e ft) l)
  | Enco.WITH(fi,t,l) -> WITH(fi,encoTy' ec pa e t,List.map (fun ft -> encoFieldTy ec pa e ft) l)
  | Enco.WHERE(fi,t,l) -> WHERE(fi,encoTy' ec pa e t,List.map (fun ft -> encoFieldTy ec pa e ft) l)
  | Enco.CHAN(fi,t) -> CHAN(fi,encoTy' ec pa e t)
  | Enco.OCHAN(fi,t) -> OCHAN(fi,encoTy' ec pa e t)
  | Enco.ICHAN(fi,t) -> ICHAN(fi,encoTy' ec pa e t)
  | Enco.RCHAN(fi,t) -> RCHAN(fi,encoTy' ec pa e t)
  | Enco.STRING(fi) -> STRING(fi)
  | Enco.BOOL(fi) -> BOOL(fi)
  | Enco.CHAR(fi) -> CHAR(fi)
  | Enco.INT(fi) -> INT(fi)
  | Enco.SITE(fi) ->
      (try match Misc.StringMap.find "TYPE_SITE" ec.Enco.forms with
	Enco.DEF_TYPE(l,newt) -> pluginTy' ec pa e newt   (* Bogus: l ignored; btw, shouldn't be Enco.enull instead of e? *)
      |	_ -> raise Not_found
      with Not_found -> SITE(fi))
  | Enco.LOC(fi) -> LOC(fi)
  | Enco.AGENTT(fi,t) -> 
      (try match Misc.StringMap.find "TYPE_AGENT" ec.Enco.forms with
	Enco.DEF_TYPE(l,newt) -> pluginTy' ec pa 
            (Enco.bind fi l (Enco.TY(t)::[]) Enco.enull) newt
      |	_ -> raise Not_found
      with Not_found -> AGENTT(fi,encoTy' ec pa e t))
  | Enco.AGENT(fi) -> 
      (try match Misc.StringMap.find "TYPE_AGENT" ec.Enco.forms with
	Enco.DEF_TYPE(l,newt) -> pluginTy' ec pa e newt  (* Bogus: l ignored *)
      |	_ -> raise Not_found
      with Not_found -> AGENT(fi))
  | Enco.DYN(fi) -> DYN(fi)
  | Enco.VART(fi,l) -> VART(fi,List.map (fun (s,t) -> (s,encoTy' ec pa e t)) l)
  | Enco.MAP(fi,t1,t2) -> MAP(fi,encoTy' ec pa e t1,encoTy' ec pa e t2)
  | Enco.ENCODE_TYPE(fi,_) -> Error.errAt fi "This construct is allowed in translation definitions only." 

and encoFieldTy ec pa e = function
    Enco.ANON(fi,t) -> ANON(fi,encoTy' ec pa e t)
  | Enco.LABELED(fi,s,t) -> LABELED(fi,s,encoTy' ec pa e t)
  | Enco.TYP(fi,id,constr) -> TY(fi,Enco.name id,encoConstr ec pa e constr)

and encoTy ec pa e = function
    None -> None
  | Some(t) -> Some(encoTy' ec pa e t)
	
and encoConstr ec pa e = function
    Enco.LT(t) -> LT(encoTy' ec pa e t)
  | Enco.EQ(t) -> EQ(encoTy' ec pa e t)

and encoConstrPluginTy ec pa e = function  (* translate types for typecase *)
    Enco.LT(t) -> LT(pluginTy' ec pa e t)
  | Enco.EQ(t) -> EQ(pluginTy' ec pa e t)

and retDec fi ec pa e dec = function
    Enco.SKIP(_) -> dec
  | p -> SEQ(fi,dec,RUN(fi,encoProc ec pa e p))

(* The top-level agent version:

let encodeToplevel ec program =
  try match Misc.StringMap.find "TOPLEVEL" ec.Enco.forms with
    Enco.DEF_DEC(l,preamble) -> 
      let toplevel_agent = Enco.fresh "$TOPLEVEL_AGENT" in
      let pa = Enco.fresh "$TRANSLATION_PARAMETERS" in 
      pluginDec ec pa
	(Enco.bind Error.unknown l 
	   (Enco.NAME(toplevel_agent)::program::[]) Enco.enull)
        (* create top-level agent *)
	(Enco.CREATE_AGENT(Error.unknown,(Error.unknown,toplevel_agent,
	 Enco.LET(Error.unknown,preamble,Enco.SKIP(Error.unknown)))::[]))
  | _ -> raise Not_found
  with
    Not_found -> Error.warning "Top-level definition not found";
      (match program with
        Enco.DEC(d) -> encoDec ec (Enco.fresh "$TRANSLATION_PARAMETERS") 
	    Enco.enull d (Enco.CONT([]))
      | _ -> Error.bug "encodeToplevel: Expected declaration")
*)


let encodeToplevel ec top =
  let (prog,pa) = 
    try match Misc.StringMap.find "PROGRAM" ec.Enco.forms with
      Enco.DEF_PROC(pa::[],prog) -> (prog,pa)
    | _ -> raise Not_found
    with
      Not_found -> (* Error.warning "Program declaration not found"; *)
      	(Enco.SKIP(Error.unknown),(Enco.fresh "$PROGRAM_PARAMETERS"))
  in
  try match Misc.StringMap.find "TOPLEVEL" ec.Enco.forms with
    Enco.DEF_DEC(l,preamble) ->
      pluginDec ec (Enco.fresh "$TRANSLATION_PARAMETERS") 
	(Enco.bind Error.unknown l 
           (Enco.PROC(prog)::Enco.NAME(pa)::[]) Enco.enull) 
	(Enco.SEQ(Error.unknown,top,preamble))
  | _ -> raise Not_found
  with
    Not_found -> (* Error.warning "Top-level declaration not found"; *)
      encoDec ec (Enco.fresh "$TRANSLATION_PARAMETERS") 
	Enco.enull (Enco.SEQ(Error.unknown,top,Enco.RUN(Error.unknown,prog))) 
	(Enco.SKIP(Error.unknown))

let encodings imports prog =
  let ec = prog.Enco.ec in
  let d = encodeToplevel ec prog.Enco.dec in
  {imports = prog.Enco.imports; static = prog.Enco.static;
   includes = prog.Enco.includes; dec = d}


