(*
 * Abstract syntax.
 *)

type info = Error.info

type proc =
  PRL of info * proc * proc
| IF of info * value * proc * proc
| INPUT of info * value * pat * proc
| OUTPUT of info * value * value

| INPA of info * pat * proc
| OUTA of info * value * value
| LDOUTA of info * value * value * value
| TRYHERE of info * value * value * proc * proc

| REP_INPUT of info * value * pat * proc
| TIMED_INPUT of info * value * value * pat * proc * proc
| REP_TIMED_INPUT of info * value * value * pat * proc * proc
| LI_SEND of info * value * value * value
| REMOTE_SEND of info * value * value * value * value
| LOCAL_SEND of info * value * value * value
| IFLOCAL of info * value * value * value * proc * proc

| LET of info * dec * proc
| SKIP of info
| TERMINATE of info
| TYPECASE of info * value * (pat * proc) list * proc
| CASE of info * ty option * value * (string * pat * proc) list 
| ENCODE_PROC of info * string list

and pat =
  RECORDp of info * fieldPat list
| ASp of info * string * ty option * pat
| VARp of info * string * ty option
| WILDp of info * ty option
| RECp of info * ty option * pat
| ENCODE_PAT of info * string list

and fieldPat =
  ANONp of info * pat
| LABELEDp of info * string * pat
| TYp of info * string * constr

and dec =
  DEF of info * (info * string * pat * proc) list
| INLINE of info * string * pat * proc
| NEW of info * string * ty
| VAL of info * pat * value
| SEQ of info * dec * dec
| RUN of info * proc
| NEWLOC of info * string
| CREATE of info * (info * string * ty * proc) list
| CREATE_AGENT of info * (info * string * proc) list
| AT of info * string
| MIGRATE of info * value
| MACRO of info * string * value
| MACROLET of info * string * string * value
| DEFER_DEC of info * dec
| INLINE_DEC of info * dec
| INLINE_VAL of info * value
| DEF_TYPE of info * string * string list * ty
| DEF_DEF of info * string * string * pat * proc
| DEF_DEC of info * string * string list * dec
| DEF_PROC of info * string * string list * proc
| ENCODE_DEC of info * string list

and ty =
  TOP of info
| TVAR of info * string
| TABS of info * (string * Kind.kind) list * ty
| REC of info * string * Kind.kind * ty
| TAPP of info * ty * ty list
| RECORD of info * fieldTy list
| WITH of info * ty * fieldTy list
| WHERE of info * ty * fieldTy list
| CHAN of info * ty
| OCHAN of info * ty
| ICHAN of info * ty
| RCHAN of info * ty
| STRING of info
| BOOL of info
| CHAR of info
| INT of info
| SITE of info
| LOC of info
| AGENTT of info * ty
| AGENT of info
| DYN of info
| VART of info * (string * ty) list
| MAP of info * ty * ty
| ENCODE_TYPE of info * string list

and constr =
  LT of ty
| EQ of ty

and fieldTy =
  ANON of info * ty
| LABELED of info * string * ty
| TY of info * string * constr

and value =
  CCODEv of info * Ccode.info * fieldVal list
| APPv of info * ty option * value * fieldVal list
| RECORDv of info * fieldVal list
| WITHv of info * ty option * value * fieldVal list
| WHEREv of info * ty option * value * fieldVal list
| STRINGv of info * string
| SITEv of info * string * int
| SELECTv of info * value * string
| VARv of info * string
| BOOLv of info * bool
| CHARv of info * char
| INTv of info * int
| LETv of info * dec * value
| RECv of info * value * ty option
| IFv of info * ty option * value * value * value
| ABSv of info * pat * proc
| DYNAMICv of info * value * ty option
| TYPECASEv of info * ty option * value * (pat * value) list * value
| VARIANTv of info * string * value
| CASEv of info * ty option * value * (string * pat * value) list
| ENCODE_VALUE of info * string list

and fieldVal =
  ANONv of info * value
| LABELEDv of info * string * value
| TYv of info * ty

type toplevel =
  {imports : (Error.info * string) list;
   static : string list;
   includes : string list;
   dec : dec}

(*
 * Extract file position information.
 *)

val procInfo: proc -> info
val patInfo: pat -> info
val valInfo: value -> info
val fvInfo: fieldVal -> info
val decInfo: dec -> info
val typeInfo: ty -> info
