(*
 * Printing fully-typed high-level syntax (O'Caml format). 
 *)

open Syntax

let caml_formatStr s =
  Format.print_string "\""; 
  Format.print_string s; 
  Format.print_string "\""

let caml_formatInfo fi = Format.print_string "info" (* Error.printInfo fi *)

let rec caml_formatConstr = function
    LT(ty) -> Format.print_string "LT("; caml_formatTy ty; Format.print_char ')'
  | EQ(ty) -> Format.print_string "EQ("; caml_formatTy ty; Format.print_char ')'

and caml_formatFieldTy = function
    ANON(fi,ty) ->
      Format.print_string "ANON(";
      caml_formatInfo fi; Format.print_char ','; caml_formatTy ty;
      Format.print_char ')'
  | LABELED(fi,s,ty) -> 
      Format.print_string "LABELED(";
      caml_formatInfo fi; Format.print_char ','; caml_formatStr s;
      Format.print_char ','; caml_formatTy ty;
      Format.print_char ')'
  | TY(fi,s,constr) -> 
      Format.print_string "TY(";
      caml_formatInfo fi; Format.print_char ','; caml_formatStr s;
      Format.print_char ','; caml_formatConstr constr;
      Format.print_char ')'
	
and caml_formatTy = function
    TOP(fi) -> 
      Format.print_string "TOP("; caml_formatInfo fi; Format.print_char ')'
  | TVAR(fi,s) -> 
      Format.print_string "TVAR(";
      caml_formatInfo fi; Format.print_char ','; caml_formatStr s;
      Format.print_char ')'
  | TABS(fi,l,ty) -> 
      let formatElem (s,k) =
  	Format.print_char '(';
  	caml_formatStr s; Format.print_char ','; Kind.caml_format k;
  	Format.print_char ')'
      in
      Format.print_string "TABS(";
      caml_formatInfo fi; Format.print_char ','; 
      Misc.formatList " [" "]" ";" formatElem l; 
      Format.print_char ','; caml_formatTy ty; Format.print_char ')'
  | REC(fi,s,k,ty) ->
      Format.print_string "REC(";
      caml_formatInfo fi; Format.print_char ','; 
      caml_formatStr s; Format.print_char ','; Kind.caml_format k;
      Format.print_char ','; caml_formatTy ty; Format.print_char ')'
  | TAPP(fi,ty,tyl) ->
      Format.print_string "TAPP(";
      caml_formatInfo fi; Format.print_char ','; 
      caml_formatTy ty; Format.print_char ','; 
      Misc.formatList " [" "]" ";" caml_formatTy tyl; 
      Format.print_char ')'
  | RECORD(fi,ftyl) ->
      Format.print_string "RECORD(";
      caml_formatInfo fi; Format.print_char ','; 
      Misc.formatList " [" "]" ";" caml_formatFieldTy ftyl;
      Format.print_char ')'
  | WITH(fi,ty,ftyl) ->
      Format.print_string "WITH(";
      caml_formatInfo fi; Format.print_char ','; caml_formatTy ty; Format.print_char ','; 
      Misc.formatList " [" "]" ";" caml_formatFieldTy ftyl;
      Format.print_char ')'
  | WHERE(fi,ty,ftyl) ->
      Format.print_string "WHERE(";
      caml_formatInfo fi; Format.print_char ','; caml_formatTy ty; Format.print_char ','; 
      Misc.formatList " [" "]" ";" caml_formatFieldTy ftyl;
      Format.print_char ')'
  | CHAN(fi,ty) -> 
      Format.print_string "CHAN(";
      caml_formatInfo fi; Format.print_char ','; caml_formatTy ty;
      Format.print_char ')'
  | OCHAN(fi,ty) -> 
      Format.print_string "OCHAN(";
      caml_formatInfo fi; Format.print_char ','; caml_formatTy ty;
      Format.print_char ')'
  | ICHAN(fi,ty) -> 
      Format.print_string "ICHAN(";
      caml_formatInfo fi; Format.print_char ','; caml_formatTy ty;
      Format.print_char ')'
  | RCHAN(fi,ty) -> 
      Format.print_string "RCHAN(";
      caml_formatInfo fi; Format.print_char ','; caml_formatTy ty;
      Format.print_char ')'
  | STRING(fi) -> 
      Format.print_string "STRING("; caml_formatInfo fi; Format.print_char ')'
  | BOOL(fi) -> 
      Format.print_string "BOOL("; caml_formatInfo fi; Format.print_char ')'
  | CHAR(fi) -> 
      Format.print_string "CHAR("; caml_formatInfo fi; Format.print_char ')'
  | INT(fi) -> 
      Format.print_string "INT("; caml_formatInfo fi; Format.print_char ')'
  | SITE(fi) -> 
      Format.print_string "SITE("; caml_formatInfo fi; Format.print_char ')'
  | LOC(fi) -> 
      Format.print_string "LOC("; caml_formatInfo fi; Format.print_char ')'
  | AGENTT(fi,ty) -> 
      Format.print_string "AGENTT(";
      caml_formatInfo fi; Format.print_char ','; caml_formatTy ty;
      Format.print_char ')'
  | AGENT(fi) -> 
      Format.print_string "AGENT("; caml_formatInfo fi; Format.print_char ')'
  | DYN(fi) -> 
      Format.print_string "DYN("; caml_formatInfo fi; Format.print_char ')'
  | VART(fi,l) ->
      let formatElem (s,ty) =
  	Format.print_char '(';
  	caml_formatStr s; Format.print_char ','; caml_formatTy ty;
  	Format.print_char ')'
      in
      Format.print_string "VART(";
      caml_formatInfo fi; Format.print_char ','; 
      Misc.formatList " [" "]" ";" formatElem l;
      Format.print_char ')'
  | MAP(fi,kty,ty) ->
      Format.print_string "MAP(";
      caml_formatInfo fi; Format.print_char ','; caml_formatTy kty; 
      Format.print_char ','; caml_formatTy ty; Format.print_char ')'
  | ENCODE_TYPE(fi,l) -> 
      Format.print_string "ENCODE_TYPE(";
      caml_formatInfo fi; Format.print_char ','; 
      Misc.formatList " [" "]" ";" caml_formatStr l;
      Format.print_char ')'

let rec caml_formatPat = function
    RECORDp(fi,l) -> 
      let caml_formatFieldPat = function
	  ANONp(fi,pat) -> 
	    Format.print_string "ANONp("; 
	    caml_formatInfo fi; Format.print_char ','; caml_formatPat pat; 
	    Format.print_char ')'
	| LABELEDp(fi,s,pat) ->
	    Format.print_string "LABELEDp("; 
	    caml_formatInfo fi; Format.print_char ','; caml_formatStr s;
	    Format.print_char ','; caml_formatPat pat; 
	    Format.print_char ')'
	| TYp(fi,s,constr) -> 
	    Format.print_string "TYp(";
	    caml_formatInfo fi; Format.print_char ','; caml_formatStr s;
	    Format.print_char ','; caml_formatConstr constr; 
	    Format.print_char ')'
      in
      Format.print_string "RECORDp("; 
      caml_formatInfo fi; Format.print_char ',';
      Misc.formatList " [" "]" ";" caml_formatFieldPat l; 
      Format.print_char ')'; 
  | ASp(fi,s,None,pat) ->
      Format.print_string "ASp("; 
      caml_formatInfo fi; Format.print_char ','; caml_formatStr s;
      Format.print_string ",None,";
      caml_formatPat pat; Format.print_char ')'
  | ASp(fi,s,Some ty,pat) ->
      Format.print_string "ASp("; 
      caml_formatInfo fi; Format.print_char ','; caml_formatStr s;
      Format.print_string ",Some "; caml_formatTy ty; 
      Format.print_char ','; caml_formatPat pat; Format.print_char ')'
  | VARp(fi,s,None) ->
      Format.print_string "VARp("; 
      caml_formatInfo fi; Format.print_char ','; caml_formatStr s;
      Format.print_string ",None"; Format.print_char ')'
  | VARp(fi,s,Some ty) ->
      Format.print_string "VARp("; 
      caml_formatInfo fi; Format.print_char ','; caml_formatStr s;
      Format.print_string ",Some "; caml_formatTy ty; 
      Format.print_char ')'
  | WILDp(fi,None) ->
      Format.print_string "WILDp("; 
      caml_formatInfo fi; Format.print_string ",None"; Format.print_char ')'
  | WILDp(fi,Some ty) ->
      Format.print_string "WILDp("; 
      caml_formatInfo fi; Format.print_string ",Some "; caml_formatTy ty; 
      Format.print_char ')'
  | RECp(fi,None,pat) ->
      Format.print_string "RECp("; 
      caml_formatInfo fi; Format.print_string ",None,";
      caml_formatPat pat; Format.print_char ')'
  | RECp(fi,Some ty,pat) ->
      Format.print_string "RECp("; 
      caml_formatInfo fi; Format.print_string ",Some "; caml_formatTy ty; 
      Format.print_char ','; caml_formatPat pat; Format.print_char ')'
  | ENCODE_PAT(fi,l) -> 
      Format.print_string "ENCODE_PAT(";
      caml_formatInfo fi; Format.print_char ','; 
      Misc.formatList " [" "]" ";" caml_formatStr l;
      Format.print_char ')'

let rec caml_formatProc = function
    PRL(fi,p,q) ->
      Format.print_string "PRL("; 
      caml_formatInfo fi; Format.print_char ',';
      Format.print_break 1 2; caml_formatBoxProc p; 
      Format.print_char ','; Format.print_break 1 2; caml_formatBoxProc q; 
      Format.print_char ')'
  | IF(fi,v,p,q) ->
      Format.print_string "IF(";
      caml_formatInfo fi; Format.print_char ',';
      Format.print_break 1 2; caml_formatVal v; 
      Format.print_char ','; Format.print_space(); Format.print_break 1 2; 
      caml_formatBoxProc p; Format.print_char ','; Format.print_space(); 
      Format.print_break 1 2; caml_formatBoxProc q;  Format.print_char ')'
  | INPUT(fi,v,pat,p) ->
      Format.print_string "INPUT("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatVal v; Format.print_char ','; 
      caml_formatPat pat; Format.print_char ','; Format.print_break 1 2; 
      caml_formatBoxProc p; Format.print_space(); Format.print_char ')'
  | OUTPUT(fi,v,av) ->
      Format.print_string "OUTPUT("; 
      caml_formatInfo fi; Format.print_char ','; 
      caml_formatVal v; Format.print_char ','; 
      caml_formatVal av; Format.print_char ')'
  | INPA(fi,pat,p) ->
      Format.print_string "INPA("; 
      caml_formatInfo fi; Format.print_char ','; 
      caml_formatPat pat; Format.print_char ','; Format.print_break 1 2; 
      caml_formatBoxProc p; Format.print_space(); Format.print_char ')'
  | OUTA(fi,v,av) ->
      Format.print_string "OUTA("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatVal v; Format.print_char ','; 
      caml_formatVal av; Format.print_char ')'
  | LDOUTA(fi,v,sv,av) ->
      Format.print_string "LDOUTA("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatVal v; Format.print_char ','; 
      caml_formatVal sv; Format.print_char ',';
      caml_formatVal av; Format.print_char ')'
  | TRYHERE(fi,v,av,p,q) -> 
      Format.print_string "TRYHERE("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatVal v; Format.print_char ','; 
      caml_formatVal av; Format.print_char ','; 
      Format.print_break 1 2; caml_formatBoxProc p; Format.print_char ','; 
      Format.print_break 1 2; caml_formatBoxProc q; Format.print_char ')'
  | REP_INPUT(fi,v,pat,p) ->
      Format.print_string "REP_INPUT("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatVal v; Format.print_char ','; 
      caml_formatPat pat; Format.print_char ','; Format.print_break 1 2; 
      caml_formatBoxProc p; Format.print_space(); Format.print_char ')'
  | TIMED_INPUT(fi,w,v,pat,p,q) -> 
      Format.print_string "TIMED_INPUT("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatVal w; Format.print_char ','; 
      caml_formatVal v; Format.print_char ','; 
      caml_formatPat pat; Format.print_char ',';
      Format.print_break 1 2; caml_formatBoxProc p; Format.print_char ','; 
      Format.print_break 1 2; caml_formatBoxProc q; Format.print_char ')'
  | REP_TIMED_INPUT(fi,w,v,pat,p,q) -> 
      Format.print_string "REP_TIMED_INPUT("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatVal w; Format.print_char ','; 
      caml_formatVal v; Format.print_char ','; 
      caml_formatPat pat; Format.print_char ',';
      Format.print_break 1 2; caml_formatBoxProc p; Format.print_char ','; 
      Format.print_break 1 2; caml_formatBoxProc q; Format.print_char ')'
  | LI_SEND(fi,v,cv,av) ->
      Format.print_string "LI_SEND("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatVal v; Format.print_char ','; 
      caml_formatVal cv; Format.print_char ',';
      caml_formatVal av; Format.print_char ')'
  | REMOTE_SEND(fi,v,sv,cv,av) ->
      Format.print_string "REMOTE_SEND("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatVal v; Format.print_char ','; 
      caml_formatVal sv; Format.print_char ','; 
      caml_formatVal cv; Format.print_char ',';
      caml_formatVal av; Format.print_char ')'
  | LOCAL_SEND(fi,v,cv,av) ->
      Format.print_string "LOCAL_SEND("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatVal v; Format.print_char ','; 
      caml_formatVal cv; Format.print_char ',';
      caml_formatVal av; Format.print_char ')'
  | IFLOCAL(fi,v,cv,av,p,q) -> 
      Format.print_string "IFLOCAL("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatVal v; Format.print_char ','; 
      caml_formatVal cv; Format.print_char ',';
      caml_formatVal av; Format.print_char ','; 
      Format.print_break 1 2; caml_formatBoxProc p; Format.print_char ','; 
      Format.print_break 1 2; caml_formatBoxProc q; Format.print_char ')'
  | LET(fi,d,p) -> 
      Format.print_string "LET("; 
      caml_formatInfo fi; Format.print_char ',';
      Format.print_break 1 2; caml_formatDec d; 
      Format.print_char ','; Format.print_break 1 2; caml_formatBoxProc p; 
      Format.print_char ')'
  | SKIP(fi) -> 
      Format.print_string "SKIP(";
      caml_formatInfo fi; Format.print_char ','; 
      Format.print_char ')'
  | TERMINATE(fi) -> 
      Format.print_string "TERMINATE(";
      caml_formatInfo fi; Format.print_char ','; 
      Format.print_char ')'
  | TYPECASE(fi,v,ts,p) -> 
      let caml_formatTCases l =
      	let caml_formatTCase (pat,p) = 
  	  Format.print_char '('; caml_formatPat pat; Format.print_char ','; 
	  Format.print_break 1 2; caml_formatBoxProc p; Format.print_char ')'
	in let rec pr needSep = function
	    [] -> ()
	  | h :: t ->
      	      if needSep then begin
	  	Format.print_char ';';
	  	Format.print_cut();
          	Format.open_hovbox 0; 
	  	caml_formatTCase h; 
	  	Format.close_box();
    	  	pr true t
      	      end 
	      else begin
	  	Format.open_hovbox 0; 
	  	caml_formatTCase h; 
	  	Format.close_box();
    	  	pr true t
    	      end
  	in
  	Format.print_string "[";
  	Format.open_hvbox 0; 
  	pr false l; 
  	Format.print_string "]";
  	Format.close_box()
      in
      Format.print_string "TYPECASE("; 
      caml_formatInfo fi; Format.print_char ','; 
      Format.print_break 1 2; caml_formatVal v; 
      Format.print_char ','; Format.print_space(); caml_formatTCases ts; 
      Format.print_char ','; Format.print_break 1 2; caml_formatBoxProc p; 
      Format.print_char ')'
  | CASE(fi,tyopt,v,cs) ->
      let caml_formatCases l =
      	let caml_formatCase (s,pat,p) = 
	  Format.print_char '(';
	  caml_formatStr s; Format.print_char ','; 
	  caml_formatPat pat; Format.print_char ','; 
	  Format.print_break 1 2; caml_formatBoxProc p; Format.print_char ')'
	in let rec pr needSep = function
	    [] -> ()
	  | h :: t ->
      	      if needSep then begin
	  	Format.print_char ';';
	  	Format.print_cut();
          	Format.open_hovbox 0; 
	  	caml_formatCase h; 
	  	Format.close_box();
    	  	pr true t
      	      end 
	      else begin
	  	Format.open_hovbox 0; 
	  	caml_formatCase h; 
	  	Format.close_box();
    	  	pr true t
    	      end
  	in
  	Format.print_string "[";
  	Format.open_hvbox 0; 
  	pr false l; 
  	Format.print_string "]";
  	Format.close_box()
      in
      Format.print_string "CASE("; 
      caml_formatInfo fi; 
      begin
      	match tyopt with
	  None -> Format.print_string ",None,";
      	| Some ty -> 
	    Format.print_string ",Some "; caml_formatTy ty; Format.print_char ',';
      end;
      Format.print_break 1 2; caml_formatVal v; 
      Format.print_char ','; Format.print_space(); caml_formatCases cs;
      Format.print_char ')'
  | ENCODE_PROC(fi,l) -> 
      Format.print_string "ENCODE_PROC(";
      caml_formatInfo fi; Format.print_char ','; 
      Misc.formatList " [" "]" ";" caml_formatStr l;
      Format.print_char ')'
	
and caml_formatBoxProc p = 
  Format.open_hvbox 0; caml_formatProc p; Format.close_box()

and caml_formatDec = function
    DEF(fi,l) -> 
      let caml_formatDefs l =
       	let caml_formatDef (fi,s,pat,p) =
  	  Format.print_char '('; 
	  caml_formatInfo fi; Format.print_char ','; 
	  caml_formatStr s; Format.print_char ','; 
	  caml_formatPat pat; Format.print_char ','; 
	  Format.print_break 1 2; caml_formatBoxProc p; Format.print_char ')'
	in let rec pr needSep = function
	    [] -> ()
	  | h :: t ->
       	      if needSep then begin
	  	Format.print_char ',';
	  	Format.print_cut();
          	Format.open_hovbox 0; 
	  	caml_formatDef h; 
	  	Format.close_box();
    	  	pr true t
      	      end 
	      else begin
	  	Format.open_hovbox 0; 
	  	caml_formatDef h; 
	  	Format.close_box();
    	  	pr true t
    	      end
  	in
  	Format.print_string "[";
  	Format.open_hvbox 0; 
  	pr false l; 
  	Format.print_string "]";
  	Format.close_box()
      in
      Format.print_string "DEF("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatDefs l; Format.print_char ')'
  | INLINE(fi,s,pat,q) ->
      Format.print_string "INLINE("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatStr s; Format.print_char ',';
      caml_formatPat pat; Format.print_char ','; Format.print_break 1 2; 
      caml_formatProc q; Format.print_space(); Format.print_char ')'
  | VAL(fi,pat,v) -> 
      Format.print_string "VAL("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatPat pat; Format.print_char ','; 
      caml_formatVal v; Format.print_char ')'
  | SEQ(fi,d1,d2) ->
      Format.print_string "SEQ("; 
      caml_formatInfo fi; Format.print_char ',';
      Format.print_break 1 2; caml_formatDec d1; 
      Format.print_char ','; Format.print_space(); caml_formatDec d2; 
      Format.print_char ')'
  | NEW(fi,s,t) -> 
      Format.print_string "NEW("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatStr s; Format.print_char ','; 
      caml_formatTy t; Format.print_char ')'
  | RUN(fi,q) -> 
      Format.print_string "RUN("; 
      caml_formatInfo fi; Format.print_char ',';
      Format.print_break 1 2; caml_formatBoxProc q; 
      Format.print_char ')'
  | NEWLOC(fi,s) ->
      Format.print_string "NEWLOC("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatStr s; Format.print_char ')'
  | CREATE(fi,l) -> 
      let caml_formatCreates l =
      	let caml_formatCreate (fi,s,ty,p) =
  	  Format.print_char '('; 
	  caml_formatInfo fi; Format.print_char ',';
	  caml_formatStr s; Format.print_char ',';
	  caml_formatTy ty; Format.print_char ','; 
	  Format.print_break 1 2; caml_formatBoxProc p; Format.print_char ')'
  	in let rec pr needSep = function
	    [] -> ()
	  | h :: t ->
       	      if needSep then begin
	  	Format.print_char ',';
	  	Format.print_cut();
          	Format.open_hovbox 0; 
	  	caml_formatCreate h; 
	  	Format.close_box();
    	  	pr true t
      	      end 
	      else begin
	  	Format.open_hovbox 0; 
	  	caml_formatCreate h; 
	  	Format.close_box();
    	  	pr true t
    	      end
  	in
  	Format.print_string "[";
  	Format.open_hvbox 0; 
  	pr false l; 
  	Format.print_string "]";
  	Format.close_box()
      in
      Format.print_string "CREATE("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatCreates l; Format.print_char ')'
  | CREATE_AGENT(fi,l) -> 
      let caml_formatCreates l =
      	let caml_formatCreate (fi,s,p) =
  	  Format.print_char '('; 
	  caml_formatInfo fi; Format.print_char ',';
	  caml_formatStr s; Format.print_char ',';
	  Format.print_break 1 2; caml_formatBoxProc p; Format.print_char ')'
  	in let rec pr needSep = function
	    [] -> ()
	  | h :: t ->
       	      if needSep then begin
	  	Format.print_char ',';
	  	Format.print_cut();
          	Format.open_hovbox 0; 
	  	caml_formatCreate h; 
	  	Format.close_box();
    	  	pr true t
      	      end 
	      else begin
	  	Format.open_hovbox 0; 
	  	caml_formatCreate h; 
	  	Format.close_box();
    	  	pr true t
    	      end
  	in
  	Format.print_string "[";
  	Format.open_hvbox 0; 
  	pr false l; 
  	Format.print_string "]";
  	Format.close_box()
      in
      Format.print_string "CREATE_AGENT("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatCreates l; Format.print_char ')'
  | AT(fi,s) -> 
      Format.print_string "AT("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatStr s; Format.print_char ')'
  | MIGRATE(fi,v) -> 
      Format.print_string "MIGRATE("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatVal v; Format.print_char ')'
  | MACRO(fi,strkey,v) -> 
      Format.print_string "MACRO("; 
      caml_formatInfo fi; Format.print_char ',';
      Format.print_string strkey; Format.print_char ',';
      caml_formatVal v; Format.print_char ')'
  | DEFER_DEC(fi,dec) -> 
      Format.print_string "DEFER_DEC(";
      caml_formatInfo fi; Format.print_char ',';
      caml_formatDec dec; Format.print_char ')'
  | INLINE_DEC(fi,dec) -> 
      Format.print_string "INLINE_DEC(";
      caml_formatInfo fi; Format.print_char ',';
      caml_formatDec dec; Format.print_char ')'
  | INLINE_VAL(fi,v) -> 
      Format.print_string "INLINE_VAL(";
      caml_formatInfo fi; Format.print_char ',';
      caml_formatVal v; Format.print_char ')'
  | DEF_TYPE(fi,s,l,t) -> 
      Format.print_string "DEF_TYPE(";
      caml_formatInfo fi; Format.print_char ',';
      caml_formatStr s; Format.print_char ',';
      Misc.formatList " [" "]" ";" caml_formatStr l; 
      Format.print_char ','; caml_formatTy t;
      Format.print_char ')'
  | DEF_DEF(fi,s,pa,pat,p) -> 
      Format.print_string "DEF_DEF(";
      caml_formatInfo fi; Format.print_char ',';
      caml_formatStr s; Format.print_char ',';
      caml_formatStr pa; Format.print_char ',';
      caml_formatPat pat; Format.print_char ',';
      caml_formatProc p; Format.print_char ')'
  | DEF_DEC(fi,s,l,d) -> 
      Format.print_string "DEF_DEC(";
      caml_formatInfo fi; Format.print_char ',';
      caml_formatStr s; Format.print_char ',';
      Misc.formatList " [" "]" ";" caml_formatStr l; 
      Format.print_char ','; caml_formatDec d;
      Format.print_char ')'
  | DEF_PROC(fi,s,l,p) -> 
      Format.print_string "DEF_PROC(";
      caml_formatInfo fi; Format.print_char ',';
      caml_formatStr s; Format.print_char ',';
      Misc.formatList " [" "]" ";" caml_formatStr l; 
      Format.print_char ','; caml_formatProc p;
      Format.print_char ')'
  | ENCODE_DEC(fi,l) -> 
      Format.print_string "ENCODE_DEC(";
      caml_formatInfo fi; Format.print_char ','; 
      Misc.formatList " [" "]" ";" caml_formatStr l;
      Format.print_char ')'
  | _ -> Error.bug "Error in tyhigh.ml: pattern-matching is not exhaustive"

and caml_formatVal = function
    CCODEv(fi,ci,l) ->
      Format.open_hovbox 0;  
      Format.print_string "CCODEv("; 
      caml_formatInfo fi; Format.print_char ',';
      Ccode.caml_formatInfo ci; Format.print_char ','; 
      Misc.formatList "[" "]" "; " caml_formatFieldVal l;
      Format.close_box(); Format.print_char ')'
  | APPv(fi,None,v,al) ->
      Format.print_string "APPv(";
      caml_formatInfo fi; Format.print_string ",None,";
      caml_formatVal v; Format.print_char ','; 
      Misc.formatList "[" "]" "; " caml_formatFieldVal al; 
      Format.print_char ')'
  | APPv(fi,Some ty,v,al) ->
      Format.print_string "APPv(";
      caml_formatInfo fi; Format.print_string ",Some "; 
      caml_formatTy ty; Format.print_char ','; 
      caml_formatVal v; Format.print_char ','; 
      Misc.formatList "[" "]" "; " caml_formatFieldVal al; 
      Format.print_char ')'
  | RECORDv(fi,l) ->
      Format.print_string "RECORDv("; 
      caml_formatInfo fi; Format.print_char ',';
      Misc.formatList "[" "]" "; " caml_formatFieldVal l; 
      Format.print_char ')'
  | WITHv(fi,None,v,l) ->
      Format.print_string "WITHv("; 
      caml_formatInfo fi; Format.print_char ',';
      Format.print_string ",None,";
      caml_formatVal v; Format.print_char ',';
      Misc.formatList "[" "]" "; " caml_formatFieldVal l; 
      Format.print_char ')'
  | WITHv(fi,Some ty,v,l) ->
      Format.print_string "WITHv("; 
      caml_formatInfo fi; Format.print_char ',';
      Format.print_string ",Some "; caml_formatTy ty; Format.print_char ',';
      caml_formatVal v; Format.print_char ',';
      Misc.formatList "[" "]" "; " caml_formatFieldVal l; 
      Format.print_char ')'
  | WHEREv(fi,None,v,l) ->
      Format.print_string "WHEREv("; 
      caml_formatInfo fi; Format.print_char ',';
      Format.print_string ",None,";
      caml_formatVal v; Format.print_char ',';
      Misc.formatList "[" "]" "; " caml_formatFieldVal l; 
      Format.print_char ')'
  | WHEREv(fi,Some ty,v,l) ->
      Format.print_string "WHEREv("; 
      caml_formatInfo fi; Format.print_char ',';
      Format.print_string ",Some "; caml_formatTy ty; Format.print_char ',';
      caml_formatVal v; Format.print_char ',';
      Misc.formatList "[" "]" "; " caml_formatFieldVal l; 
      Format.print_char ')'
  | STRINGv(fi,s) ->
      Format.print_string "STRINGv("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatStr s; 
      Format.print_string ")"
  | SITEv(fi,s,i) ->
      Format.print_string "SITEv("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatStr s; Format.print_int i;
      Format.print_string ")"
  | SELECTv(fi,v,s) ->
      Format.print_string "SELECTv("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatVal v; Format.print_char ','; 
      caml_formatStr s; Format.print_char ')'
  | VARv(fi,s) ->  
      Format.print_string "VARv("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatStr s; Format.print_char ')'
  | BOOLv(fi,b) ->
      Format.print_string "BOOLv("; 
      caml_formatInfo fi; Format.print_char ',';
      Format.print_bool b; Format.print_char ')'
  | CHARv(fi,x) ->
      Format.print_string "CHARv(";
      caml_formatInfo fi;
      Format.print_string ",\'"; Format.print_char x; 
      Format.print_string "\')";
  | INTv(fi,k) ->
      Format.print_string "INTv("; 
      caml_formatInfo fi; Format.print_char ',';
      Format.print_int k; Format.print_char ')'
  | LETv(fi,d,v) ->
      Format.print_string "LETv("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatDec d; Format.print_char ','; 
      caml_formatVal v; Format.print_char ')'
  | RECv(fi,v,None) ->
      Format.print_string "RECv("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatVal v;
      Format.print_string ",None)";
  | RECv(fi,v,Some ty) ->
      Format.print_string "RECv("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatVal v;
      Format.print_string ",Some "; caml_formatTy ty; Format.print_char ')'
  | IFv(fi,None,b,v1,v2) ->
      Format.print_string "IFv("; 
      caml_formatInfo fi; 
      Format.print_string ",None,";
      caml_formatVal b; Format.print_char ','; 
      caml_formatVal v1; Format.print_char ','; caml_formatVal v2; 
      Format.print_char ')'
  | IFv(fi,Some ty,b,v1,v2) ->
      Format.print_string "IFv("; 
      caml_formatInfo fi; 
      Format.print_string ",Some "; caml_formatTy ty; Format.print_char ',';
      caml_formatVal b; Format.print_char ','; 
      caml_formatVal v1; Format.print_char ','; caml_formatVal v2; 
      Format.print_char ')'
  | ABSv(fi,pat,p) ->
      Format.print_string "ABSv("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatPat pat; Format.print_char ','; 
      Format.print_break 1 2; caml_formatBoxProc p; Format.print_char ')'
  | DYNAMICv(fi,v,None) ->
      Format.print_string "DYNAMICv("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatVal v; Format.print_string ",None)"
  | DYNAMICv(fi,v,Some ty) ->
      Format.print_string "DYNAMICv("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatVal v; Format.print_string ",Some "; caml_formatTy ty; 
      Format.print_char ')'
  | TYPECASEv(fi, None, v, ts, velse) ->
      let caml_formatTCasev (pat,v) = 
	Format.print_char '('; caml_formatPat pat; Format.print_char ','; 
	caml_formatVal v; Format.print_char ')' 
      in
      Format.print_string "TYPECASEv("; 
      caml_formatInfo fi; Format.print_string ",None,";
      caml_formatVal v; Format. print_char ','; 
      Misc.formatList "[" "]" "; " caml_formatTCasev ts;
      Format.print_char ','; caml_formatVal velse ; Format.print_char ')'
  | TYPECASEv(fi, Some ty, v, ts, velse) ->
      let caml_formatTCasev (pat,v) = 
	Format.print_char '('; caml_formatPat pat; Format.print_char ','; 
	caml_formatVal v; Format.print_char ')' 
      in
      Format.print_string "TYPECASEv("; 
      caml_formatInfo fi; 
      Format.print_string ",Some "; caml_formatTy ty; Format. print_char ',';
      caml_formatVal v; Format. print_char ','; 
      Misc.formatList "[" "]" "; " caml_formatTCasev ts;
      Format.print_char ','; caml_formatVal velse ; Format.print_char ')'
  | VARIANTv(fi,s,v) ->
      Format.print_string "VARIANTv("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatStr s; Format.print_char ',';
      caml_formatVal v; Format.print_char ')'
  | CASEv(fi, None, v, cs) ->
      let caml_formatCasev (s, pat,v) = 
	Format.print_char '('; 
	caml_formatStr s; Format.print_char ','; 
	caml_formatPat pat; Format.print_char ','; caml_formatVal v; 
	Format.print_char ')' 
      in
      Format.print_string "CASEv("; 
      caml_formatInfo fi; Format.print_string ",None,";
      caml_formatVal v; Format. print_char ','; 
      Misc.formatList "[" "]" "; " caml_formatCasev cs; 
      Format.print_char ')'
  | CASEv(fi, Some ty, v, cs) ->
      let caml_formatCasev (s, pat,v) = 
	Format.print_char '('; 
	caml_formatStr s; Format.print_char ','; 
	caml_formatPat pat; Format.print_char ','; caml_formatVal v; 
	Format.print_char ')' 
      in
      Format.print_string "CASEv("; 
      caml_formatInfo fi; 
      Format.print_string ",Some "; caml_formatTy ty; Format. print_char ',';
      caml_formatVal v; Format. print_char ','; 
      Misc.formatList "[" "]" "; " caml_formatCasev cs; 
      Format.print_char ')'
  | ENCODE_VALUE(fi,l) -> 
      Format.print_string "ENCODE_VALUE(";
      caml_formatInfo fi; Format.print_char ','; 
      Misc.formatList " [" "]" ";" caml_formatStr l;
      Format.print_char ')'

and caml_formatFieldVal = function
    ANONv(fi,v) -> 
      Format.print_string "ANONv("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatVal v; Format.print_char ')'
  | LABELEDv(fi,s,v) ->
      Format.print_string "LABELEDv("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatStr s; Format.print_char ',';
      caml_formatVal v; Format.print_char ')'
  | TYv(fi,ty) -> 
      Format.print_string "TYv("; 
      caml_formatInfo fi; Format.print_char ',';
      caml_formatTy ty; Format.print_char ')'

let printToplevel s = caml_formatStr s; Format.print_newline()

let caml_format prog =
  Format.print_string "{- Includes -}";
  Format.print_newline(); List.iter printToplevel prog.includes;
  Format.print_string "{- Static C Code -}";
  Format.print_newline(); List.iter printToplevel prog.static;
  Format.print_string "{- Processes -}";
  Format.print_newline(); caml_formatDec prog.dec; 
  Format.print_newline()
